
from threading import Thread, RLock
from Queue import Queue, Empty
from traceback import print_exc, format_exc
from schema import default as default_widget_schema
import weakref
import default_features
import libautobus
from concurrent import synchronized
from schema.categories import TOPLEVEL, CONTAINER, WIDGET
import textwrap
import pydoc

"""
RTK is a remote widget library. It stands for Remote ToolKit, and it's similar
to X, with two differences: it reverses the roles of client and server used in
X, and it sends specific instructions about which widgets to create and how to
lay them out instead of lower-level graphics instructions. The result is that
RTK applications look native to the operating system that the viewer is run
on.

Since an RTK application is a server instead of a client, multiple users can
connect to the same application at once. Each one receives a different copy
of the application, but they all run in the same process so they can
inter-communicate. (With the threading protocol, each copy of the application
runs in its own thread, while with the async protocol, a single thread manages
them all and Python's asyncore library is used to manage communication with
the viewers.)

A reference implementation viewer is provided in the form of the rtkinter
module. This viewer uses Tkinter to display the application.

RTK is thread-safe. Multiple threads can access and modify RTK widgets
concurrently. Instances of an application can even pass connections around and
modify widgets on other connections in the same process.

Various RTK examples are provided in the examples package. A simple
"Hello, world" application that shows a window titled "Hello" containing one
label whose text is "Hello, world!" could be written thus:

from librtk.protocols import ThreadedServer
def start_app(connection):
    w = connection.Window(title="Hello")
    w.close_request.listen(connection.close)
    l = connection.Label(w, text="Hello, world!")
server = ThreadedServer("", 6785, start_app)
server.start()

A viewer would then be started and connected to port 6785. rtkinter could be
used to view this application thus:

./run rtkinter rtk://localhost:6785

This can be run multiple times simultaneously, or run from a different machine
than the one that the application is running on.

If you want to experiment around with a single viewer from the Python REPL,
you can use the listen utility function in librtk to listen for one viewer to
connect and return the connection. This can be used like so:

from librtk import listen
connection = listen(6785) # This will return once you start up a viewer,
                          # such as rtkinter, pointing at it
w = connection.Window(title="Hello")
...etc...

Updates will be sent to the viewer immediately as each line is executed. So as
soon as you hit enter after typing connection.Window(title="Hello"), the
window will pop open in the viewer.

dir() can be used on the connection to find out what widgets are available.

To get help with a particular widget from the command line, use
some_widget.help(). This will open an interactive help console similar to
Python's built-in help function that shows information about that widget and
what attributes it supports. This function is available on both widget
constructors (connection.Window.help(), for example) and widget instances
(w = connection.Window(...); w.help(), for example). Since the help function
derives its information solely from the widget schema (which is available as a
field on the connection object), it should be possible to write an HTML
documentation generator for this; this is something I plan on doing in the
future.

RTK can also be used as a traditional widget toolkit by using... TODO: finish 
"""

from afn.utils.listener import Event

global_lock = RLock()
text_doc = pydoc.TextDoc()
locked = synchronized(global_lock)

class MissingFeatures(Exception):
    """
    Thrown by a validator if it decides that the client doesn't have enough
    features for the program to sanely run on it.
    """
    pass

class EventThread(Thread):
    """
    An event dispatch thread. You create one of these by passing in an
    instance of Queue (or any other object that has a no-argument get function)
    and then start it. It reads callable objects off of the queue and calls
    them until the item returned from the queue is None.
    """
    def __init__(self, queue=None):
        """
        Creates a new EventThread with the specified queue. If no queue is
        specified, a new one is created. The queue is then stored in the
        queue attribute on the event thread.
        
        This doesn't start the event thread; you need to remember to call
        start on the event thread to start it.
        """
        Thread.__init__(self)
        if queue is None:
            queue = Queue()
        self.queue = queue
    
    def run(self):
        item = self.queue.get()
        while item is not None:
            try:
                item()
            except:
                print_exc()
            item = self.queue.get()
    
    def schedule(self, function):
        """
        Same as self.queue.put(function).
        """
        self.queue.put(function)


class Connection(object):
    @locked
    def __init__(self, protocol, connect_function, validators=[],
            use_default_widgets=True, custom_features=[]):
        self.protocol = protocol
        self.connect_function = connect_function
        self.validators = validators[:]
        if use_default_widgets:
            self.validators.append(default_validator)
        self.children = []
        self.widgets = {}
        self.schema = {}
        self.combined_class_cache = weakref.WeakValueDictionary()
        self.layout_classes = {}
        self.features = default_features.features + custom_features
        self.started = False
        self.handshake_finished = False
        self.closed = False
        self.pre_start_messages = []
        self.close_functions = []
        self.client_features = None
        protocol.protocol_init(self)
    
    @locked
    def add_close_function(self, function):
        """
        Adds a function that will be called when this connection is closed.
        """
        self.close_functions.append(function)
    
    @locked
    def start(self):
        """
        Starts this connection. This should be called after the connection has
        been set up (all validators have been added with add_validator, etc).
        """
        if self.started:
            raise Exception("You can't call Connection.start twice.")
        self.started = True
        self.protocol.protocol_start()
        # If any messages were buffered up by self.protocol_receive before
        # we started, we'll process them now.
        messages = self.pre_start_messages
        self.pre_start_messages = None
        for message in messages:
            self.protocol_receive(message)
    
    @locked
    def send(self, packet):
        """
        Used by internal RTK code to send a packet to the client. The packet
        should be a dict ready to be put into json.dumps.
        """
        self.protocol.protocol_send(packet)
    
    @locked
    def add_validator(self, validator):
        """
        Adds a new validator function. TODO: more on this later.
        """
        self.validators.append(validator)
    
    @locked
    def schedule(self, function):
        """
        Schedules a function to be run as soon as possible. The function will
        be put on the event queue, so no incoming changes will be processed
        while the function is running.
        """
        self.protocol.protocol_event_ready(function)
    
    @locked
    def fatal_error(self, message):
        # TODO: Right now, errors that are most likely the fault of the
        # programmer and errors that result from the client messing up are
        # both printed by this function. Perhaps split them out so the
        # application developer knows what to look for to know if they have
        # potential problems with their application.
        print "FATAL CONNECTION ERROR: " + message
        self.send({"action": "drop", "text": message})
        self.close(True)
    
    @locked
    def protocol_receive(self, data):
        """
        Called by subclasses of Connection when they have new data for
        Connection to process. data should be the return value of json.loads
        or a compatible object.
        """
        if not self.started: # Haven't started yet, so buffer up messages
            # until when someone calls self.start()
            self.pre_start_messages.append(data)
            return
        if not self.handshake_finished: # Perform handshake
            if not data["action"] == "connect":
                self.fatal_error("First message must be a connect message")
                return
            features = data["features"]
            self.client_features = features
            application = data.get("application", None)
            self.application = application
            try:
                for validator in self.validators:
                    validator(features, self.schema, self)
            except MissingFeatures, e:
                self.fatal_error("Missing features: " + str(e))
                return
            self.send({"action": "accept", "features": self.features})
            self.handshake_finished = True
            self.create_widget_constructors()
            try:
                self.connect_function(self)
            except:
                print_exc()
                self.fatal_error("Exception in connect function")
                return
            return
        # If this is directed to a widget, we'll look it up and dispatch it.
        # If it's not, we'll ignore it for now and issue a message about it,
        # since there aren't any messages at present that don't target widgets.
        #
        # We'll do the actual dispatching on the event queue so that widgets
        # aren't modified while a listener is running. 
        if "id" in data:
            def dispatch():
                widget = self.widgets.get(data["id"], None)
                if widget is None: # Ignore messages targeted to nonexistent
                    # widgets; most likely the widget was deleted just before
                    # the event arrived across the wire.
                    return
                # Dispatch the message.
                widget.dispatch(data)
            self.schedule(dispatch)
        elif data["action"] == "error":
            print "CLIENT REPORTED AN ERROR: " + data["text"]
        elif data["action"] == "drop":
            print "CLIENT REPORTED A FATAL ERROR: " + data["text"]
            self.close(hard=True)
        else:
            print ("Message (with action " + data["action"] + ") ignored "
                    "because it does not have a target id")
    
    @locked
    def protocol_connection_lost(self):
        """
        Called by subclasses of Connection when the connection has been lost.
        If Connection previously called protocol_close (which subclasses
        themselves implement), the connection is not required (but is allowed)
        to call this function. If protocol_close has not been called, this
        will cause it to be called.
        """
        self.close(True)
    
    @locked
    def close(self, hard=False):
        """
        Closes this connection. If hard is True, a hard close is performed,
        which essentially calls protocol_close without doing anything else. If
        hard is False (the default), various cleanup actions (such as
        destroying all of this connection's widgets) are performed first.
        
        If the connection has already been closed, this does nothing. 
        """
        if self.closed:
            return
        self.closed = True
        if not hard:
            for widget in self.children[:]:
                widget.destroy()
        self.protocol.protocol_close()
        for function in self.close_functions:
            try:
                function()
            except:
                print_exc()
    
    def add_child(self, child):
        self.children.append(child)
        child.send_create()
    
    def get_child_index(self, child):
        return self.children.index(child)
    
    def create_widget_constructors(self):
        """
        Called by internal code to create self.Window, self.Button, etc.
        These are generated from the widget schema. External code should not
        use this.
        """
        for name, value in self.schema.items():
            if value.type == TOPLEVEL:
                setattr(self, name, ResidentWidgetConstructor(value, value[0], self))
            else:
                setattr(self, name, ResidentWidgetConstructor(value, value[0]))
    
    def __getitem__(self, item):
        return self.children[item]
    
    def __len__(self):
        return len(self.children)
    
    def __nonzero__(self):
        return True
    
class ResidentWidget(object):
    """
    A client-side widget. Instances of this class correspond to specific
    widgets being used on the client side.
    """
    
    @locked
    def __init__(self, type, parent, **kwargs):
        object.__setattr__(self, "_resident_ready", False)
        self.type = type
        self.id = libautobus.get_next_id()
        if isinstance(parent, Connection):
            self.parent = None
            self.connection = parent
        else:
            self.parent = parent
            self.connection = parent.connection
        self._schema = self.connection.schema[type]
        self.children = [] # Create even for non-container types to make some
        # iteration stuff simpler. This will just be empty for such types.
        self.server_name, self.doc, self.category = self._schema[0:3]
        if self.parent is None and self.category != TOPLEVEL:
            raise Exception("Containers and widgets must have a container or "
                    "a toplevel specified as the parent when creating the "
                    "widget.")
        if self.parent is not None and self.category == TOPLEVEL:
            raise Exception("Toplevels must not have a parent specified (or "
                    "the parent must be a connection).")
        if self.category != TOPLEVEL and self.parent.category == WIDGET:
            raise Exception("Containers and widgets can only be added to "
                    "toplevels and containers, not other widgets.")
        (self.widget_schema, self.layout_schema, self.state_schema,
        self.call_schema, self.event_schema) = self._schema[3:8]
        for key, (name, doc, writable, default) in self.widget_schema.items():
            if default is None and key not in kwargs:
                raise Exception("Widget property " + key + " is required to "
                        "construct a widget of type " + type + ", but this "
                        "property was unspecified.")
        if self.parent is not None:
            self.parent.validate_layout_attributes(kwargs)
        # At this point we've validated everything, so we can do ahead and
        # start setting stuff up.
        self.connection.widgets[self.id] = self
        self.widget_properties = dict([(k, kwargs.get(k, d)) 
                for k, (n, doc, w, d) in self.widget_schema.items()])
        if self.parent is not None:
            self.layout_properties = dict([(k, kwargs.get(k, d)) 
                    for k, (n, doc, w, d) in self.parent.layout_schema.items()])
        else:
            self.layout_properties = {}
        self.state_properties = dict([(k, d)
                for k, (n, doc, d,) in self.state_schema.items()])
        self.state_events = dict([(k, Event()) for k in self.state_schema.keys()])
        self.events = dict([(k, Event()) for k in self.event_schema.keys()])
        for k in self.events:
            if k in kwargs:
                self.events[k].listen(kwargs[k])
        self.owner.add_child(self)
        self._resident_ready = True
    
    def help(self):
        """
        Shows help for this widget. This is quite similar to Python's built-in
        help function, but the format is specific to RTK widgets.
        """
        show_help(self._schema, True, self.parent._schema
                if self.parent is not None else None)
    
    @property # Doesn't need to be locked since self.parent and self.connection
    # never change after construction
    def owner(self):
        """
        The owner of this widget. This is the same as the widget's parent, or
        the connection if this widget is a toplevel widget.
        """
        return self.parent or self.connection
    
    @locked
    def validate_layout_attributes(self, attributes):
        """
        Validates that the layout attributes required by this widget, which
        must be a container or a toplevel, are present in the specified map of
        attributes.
        
        This is used by ResidentWidget for a child widget to validate that it
        contains all of the layout attributes needed by its parent-to-be.
        Outside code generally shouldn't call this.
        """
        for key, (name, doc, writable, default) in self.layout_schema.items():
            if default is None and key not in attributes:
                raise Exception("Layout property " + key + " is required for "
                        + "children of " + self.category + "s of type "
                         + self.type + ", but this property was unspecified.")
    
    @locked
    def add_child(self, child):
        """
        Requests that this container or toplevel adopt the specified child.
        This class implements this to add the child to the list of child
        widgets and call the child's send_add function.
        
        This generally should not be called by outside code. It's called
        implicitly when constructing a ResidentWidget instance, so you don't
        need to call it yourself.
        """
        self.children.append(child)
        child.send_create()
    
    @locked
    def send_create(self):
        """
        Sends a message to the client telling it to perform the actual
        creation of this widget. Outside code generally shouldn't call this;
        it's called by a widget's parent-to-be in add_child.
        """
        message = {"action": "create", "id": self.id, "type": self.type,
                "index": self.owner.get_child_index(self), "p_widget":
                self.widget_properties, "p_layout": self.layout_properties}
        if self.parent is not None:
            message["parent"] = self.parent.id
        self.connection.send(message)
    
    @locked
    def get_child_index(self, child):
        """
        Returns the index at which a particular child is located in this
        container or toplevel. If the specified widget is not a child of this
        widget, ValueError will be raised.
        """
        return self.children.index(child)
    
    @locked
    def dispatch(self, message):
        """
        Dispatches an event or a state property change to this widget. This
        must only be called on the event queue. This will take care of
        notifying any interested listeners.
        """
        if message["action"] == "set_state":
            properties = message["properties"]
            self.state_properties.update(properties)
            for k in properties.keys():
                self.state_events[k]()
        elif message["action"] == "event":
            self.events[message["name"]](*message["args"])
    
    @locked
    def _get_widget(self, name):
        return self.widget_properties[name]
    
    @locked
    def _get_layout(self, name):
        return self.layout_properties[name]
    
    @locked
    def _get_state(self, name):
        return self.state_properties[name]
    
    @locked
    def _get_call(self, name):
        return ResidentCall(self, name)
    
    @locked
    def _get_event(self, name):
        return self.events[name]
    
    @locked
    def _get_state_event(self, name):
        """
        Name should be the name of the state property without "_changed" on
        the end
        """
        return self.state_events[name]
    
    @locked
    def __getattr__(self, name):
        if not object.__getattribute__(self, "_resident_ready"):
            return object.__getattribute__(self, name)
        if name in self.widget_schema:
            return self._get_widget(name)
        elif self.parent is not None and name in self.parent.layout_schema:
            return self._get_layout(name)
        elif name in self.state_schema:
            return self._get_state(name)
        elif name in self.call_schema:
            return self._get_call(name)
        elif name in self.event_schema:
            return self._get_event(name)
        elif (name.endswith("_changed") and
                name[:-len("_changed")] in self.state_schema):
            return self._get_state_event(name[:-len("_changed")])
        # We don't need to delegate to object since __getattr__ is only
        # called if Python can't find the attribute any other way
        raise AttributeError("Widget of type " + self.type + 
                " has no property " + name)
    
    @locked
    def _set_widget(self, name, value):
        if not self.widget_schema[name].writable:
            raise Exception("You can't modify the widget property " + name
                    + " on a widget of type " + self.type + ". This property "
                    "can only be specified when the widget is created.")
        self.widget_properties[name] = value
        self.send_set_widget(name)
    
    @locked
    def _set_layout(self, name, value):
        if not self.parent.layout_schema[name].writable:
            raise Exception("You can't modify the layout property " + name
                    + " on a widget of type " + self.type + " contained "
                    " in a parent of type " + self.parent.type + ". This "
                    "property can only be specified when the widget is created.")
        self.layout_properties[name] = value
        self.parent.send_set_layout(self, name)
    
    @locked
    def _set_state(self, name, value):
        raise Exception("You can't modify the state property " + name
                + " on a widget of type " + self.type + ". State "
                "properties can only be modified by the client.")
    
    @locked
    def _set_call(self, name, value):
        raise Exception("You can't set values for properties "
                "corresponding to calls on widgets. You just tried to "
                "set the property " + name + " on a widget of type "
                + self.type + ".")
    
    @locked
    def _set_event(self, name, value):
        raise Exception("Events can't be set for now. In the future, this "
                "will be allowed, with the result that all listeners on "
                "the specified event will be removed and replaced with "
                "the value you're assigning to this property. You just "
                "tried to set the value for the event " + name + " on a "
                "wiget of type " + self.type + ".")
    
    @locked
    def _set_state_event(self, name, value):
        """
        Name should be the name of the state property without "_changed" on
        the end
        """
        raise Exception("State events can't be set for now. In the future, this "
                "will be allowed, with the result that all listeners on "
                "the specified state event will be removed and replaced with "
                "the value you're assigning to this property. You just "
                "tried to set the value for the state event " + name + "_changed "
                "on a widget of type " + self.type + ".")
    
    @locked
    def __setattr__(self, name, value):
        if not object.__getattribute__(self, "_resident_ready"):
            object.__setattr__(self, name, value)
            return
        if name in self.widget_schema:
            self._set_widget(name, value)
        elif self.parent is not None and name in self.parent.layout_schema:
            self._set_layout(name, value)
        elif name in self.state_schema:
            raise Exception("You can't modify the state property " + name
                    + " on a widget of type " + self.type + ". State "
                    "attributes can only be modified by the client.")
        elif name in self.call_schema:
            self._set_call(name, value)
        elif name in self.event_schema:
            self._set_event(name, value)
        elif (name.endswith("_changed") and
                name[:-len("_changed")] in self.state_schema):
            self._set_state_event(name[:-len("_changed")], value)
        else:
            object.__setattr__(self, name, value)
    
    @locked
    def send_set_widget(self, name):
        self.connection.send({"action": "set_widget", "id": self.id, "properties":
                {name: getattr(self, name)}})
    
    @locked
    def send_set_layout(self, child, name):
        self.connection.send({"action": "set_layout", "id": child.id, "properties":
                {name: getattr(child, name)}})
    
    @locked
    def destroy(self):
        for child in self.children[:]:
            child.destroy()
        self.connection.send({"action": "destroy", "id": self.id})
        # TODO: perhaps route this through the parent
        self.owner.children.remove(self)
        del self.connection.widgets[self.id]
    
    def __getitem__(self, item):
        # TODO: in the future, perhaps allow this as a shortcut for getattr
        # for string items
        return self.children[item]
    
    def __len__(self):
        return len(self.children)
    
    def __nonzero__(self):
        return True
    
    def __str__(self):
        if self.parent is not None:
            in_string = (" with parent " + self.parent.type + " "
                    + str(self.parent.id))
        else:
            in_string = "" 
        return "<" + self.type + " instance " + str(self.id) + in_string + ">"
    
    __repr__ = __str__

class ResidentCall(object):
    def __init__(self, widget, name):
        self.widget = widget
        self.name = name
    
    def __call__(self, *args):
        self.widget.connection.send({"action": "call", "id": self.widget.id,
                "name": self.name, "args": args})
    
    def __repr__(self):
        return ("<Call " + self.widget.type + "." + self.name + " on " +
                str(self.widget.id) + ">")
    
    __str__ = __repr__


def gen_attribute_doc(attribute):
    """
    Generates documentation for a single attribute on a widget. attribute
    should be an instance of WidgetPropertySchema, LayoutPropertySchema,
    StatePropertySchema, CallSchema, or EventSchema, in
    librtk.schema.structure.
    """
    return (text_doc.bold(attribute.name) + "\n" + 
            text_doc.indent(linewrap(attribute.doc), "    "))

def gen_group_doc(map, title, add_suffix=True):
    """
    Generates documentation for a particular group of attributes on a widget.
    The group should be specified as a dictionary, and is typically the
    value of one of the attributes on a WidgetSchema instance such as
    widget_properties. The title defines what we're talking about; for
    example, it could be "Widget properties".
    
    Normally, the title will have " defined here:" automatically appended to
    the end of it. Specifying add_suffix=False will disable this.
    
    If the specified map is empty, the empty string will be returned.
    """
    if len(map) == 0:
        return ""
    doc = title
    if add_suffix:
        doc += " defined here:"
    doc += "\n\n"
    keys = list(map.keys())
    keys.sort()
    doc += "\n\n".join(gen_attribute_doc(map[key]) for key in keys)
    return doc

def gen_widget_doc(schema, instance=False, parent_schema=None):
    """
    Generates documentation for the specified widget schema. If instance is
    False, the widget's layout properties will be included under a section
    titled "Layout properties defined here". If instance is True, the
    widget's layout properties will not be included; the specified parent's
    layout properties will be included instead, and they will be included
    under a section titled "Layout properties specified by parent container
    ParentType", where ParentType is the type of the parent and "container"
    is "toplevel" instead if the parent is a toplevel.
    """
    doc = schema.type + " " + text_doc.bold(schema.name) + "\n"
    child_doc = linewrap(schema.doc, 74)
    group_docs = [
            gen_group_doc(schema.widget_properties, "Widget properties"),
            (
                None
                if instance and parent_schema is None else
                gen_group_doc(parent_schema.layout_properties,
                        "Layout properties specified by parent " +
                        parent_schema.type + " " + parent_schema.name + ":",
                        add_suffix=False)
                if instance else
                gen_group_doc(schema.layout_properties, "Layout properties")
            ),
            gen_group_doc(schema.state_properties, "State properties"),
            gen_group_doc(schema.calls, "Calls"),
            gen_group_doc(schema.events, "Events")]
    group_docs = filter(None, group_docs)
    if group_docs:
        child_doc += "\n\n"
        child_doc += ("\n\n" + "-" * 70 + "\n").join(group_docs)
    doc += text_doc.indent(child_doc, " |  ")
    return doc

def show_help(schema, instance=False, parent_schema=None):
    pydoc.pager(gen_widget_doc(schema, instance, parent_schema))

def linewrap(text, width=70):
    return "\n\n".join([textwrap.fill(x, width) for x in text.split("\n\n")])

class ResidentWidgetConstructor(object):
    def __init__(self, schema, type, default_parent=None):
        self.schema = schema
        self.type = type
        self.default_parent = default_parent
    
    def help(self):
        show_help(self.schema)
    
    def __call__(self, parent=None, **kwargs):
        if parent is not None and self.default_parent is not None:
            raise Exception("Can't specify explicit parent for a constructor "
                    "with a default parent specified")
        if parent is None:
            parent = self.default_parent
        if parent is None:
            raise Exception("Need to specify the parent to create this "
                    "widget under, since it's not a toplevel")
        return ResidentWidget(self.type, parent, **kwargs)
    
    def __str__(self):
        return "<" + self.type + " constructor>"
    
    __repr__ = __str__


def default_validator(features, schema, connection):
    schema.update(default_widget_schema.schema)

def listen(port=6785, localhost_only=True, handshake_timeout=10):
    """
    Listens on the specified port for a single connection from a single
    viewer. The connection's handshake is then performed and the connection
    object returned.
    
    If localhost_only is True (the default), only connections on the loopback
    adapter will be allowed. If it's False, connections on any interface will
    be allowed.
    
    handshake_timeout is the number of seconds to wait after a connection is
    received for the handshake to be performed by the client. This defaults to
    10 seconds. If the handshake is not performed within this many seconds,
    the connection will be closed and an exception thrown. This does not
    affect actually listening for a connection before one is established;
    this function will wait indefinitely for this to happen.
    """
    from protocols import ThreadedProtocol
    import socket
    server = socket.socket()
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(("127.0.0.1" if localhost_only else "", port))
    server.listen(1)
    client, _ = server.accept()
    server.close()
    protocol = ThreadedProtocol(client)
    queue = Queue()
    def connected(connection):
        queue.put(None)
    connection = Connection(protocol, connected)
    connection.start()
    try:
        queue.get(block=True, timeout=handshake_timeout)
    except Empty:
        try:
            connection.close()
        except:
            pass
        raise Exception("Handshake not completed within the specified number "
                "of seconds (which was " + str(handshake_timeout) + ")")
    return connection












































class AutobusException(Exception):
    pass


class TimeoutException(AutobusException):
    pass


class ConnectionException(AutobusException):
    pass


class ConnectionLostException(ConnectionException):
    pass


class NotConnectedException(ConnectionException):
    pass


class InvalidValueException(AutobusException):
    pass


class RemoteUserException(AutobusException):
    pass


class CommandErrorException(AutobusException):
    pass


class NullMessageException(AutobusException):
    pass


class QueryCanceledException(AutobusException):
    pass


class NoMatchingServiceException(AutobusException):
    pass


class ClosedException(AutobusException):
    pass






from libautobus import AutobusConnection
from threading import Thread

server = AutobusConnection()
server.connect()

def get_multiple(index):
    print "Starting for " + str(index) + "..."
    interface = server["afntest.autobus3." + str(index)]
    for i in range(60):
        interface.say_hello("Alex")
    print "Done"
threads = []
for i in range(50):
    threads.append(Thread(target=get_multiple, args=(i%10,)))
    threads[-1].start()

print "Running"
for thread in threads:
    thread.join()
print "Shutting down"
server.shutdown()




import gobject
from gobject import idle_add
import gtk
from concurrent import as_new_thread

"""
A module that knows how to show windows browsing timer set models. Instances
of TimerSetDisplay can be created, passing in a controller object from which
the model will be obtained and to which all timer modifications by the user
will be sent.

When using an instance of TimerSetDisplay to view a timer server connected over
Autobus, the Autobus proxy representing the interface registered by the timer
server should be passed in as the controller object. TimerSetDisplay will take
care of everything else. Due to a current bug in libautobus, the proxy must be
obtained and the TimerSetDisplay constructed *before* calling the connection's
connect() or start_connecting() functions for the first time.

NOTE: Applications using this class must import gobject and call
gobject.threads_init() before creating instances of this class. Failure to do
so will result in Python deadlocking.
"""

class TimerSetDisplay(object):
    def __init__(self, title, controller, prefs_function):
        """
        Creates a new TimerSetDisplay whose window will use the specified
        title. The display will also use the specified cotroller object.
        When the preferences button on the toolbar is clicked, the specified
        preferences function will be called in order to let the application
        itself deal with preferences.
        """
        self.title = title
        self.controller = controller
        self.prefs_function = prefs_function
        self.builder = gtk.Builder()
        self.builder.add_from_file("glade/traytimer/traytimer.xml")
        self.builder.connect_signals(self)
        self.window = self.builder.get_object("window")
        self.tabs = self.builder.get_object("tabs")
        controller["object":"timers"].watch(self.timers_object_changed)
        controller["object":"startup"].watch(self.startup_object_changed)
    
    @as_new_thread
    def on_timer_add(self, button):
        number = self.controller.create()
        idle_add(self.switch_to_tab, number)
    
    @as_new_thread
    def on_timer_delete(self, button):
        timer = self.get_current_timer()
        dialog = gtk.MessageDialog(self.window, type=gtk.MESSAGE_QUESTION,
                buttons=gtk.BUTTONS_YES_NO, message_format="Are you sure you "
                "want to delete timer " + str(timer) + "?")
        response = dialog.run()
        dialog.destroy()
        if response == gtk.RESPONSE_YES:
            self.controller.delete(timer)
    
    def on_timer_announce(self, button):
        self.controller.announce(self.get_current_timer())
    
    def on_timer_rename(self, button):
        pass
    
    def on_preferences(self, button):
        self.prefs_function()
    
    def switch_to_tab(self, timer_number):
        """
        Must be called on the GTK thread.
        """
        for p in self.tabs.get_n_pages():
            label = self.tabs.get_tab_label_text(self.tabs.get_nth_tab(p))
            if self.extract_number_from_name(label) == timer_number:
                self.tabs.set_current_page(p)
                return
        print "Switch to timer " + str(timer_number) + ": no such timer"
    
    def get_current_timer(self):
        """
        Must be called on the GTK thread.
        """
        page_number = self.tabs.get_current_page()
        if page_number == -1:
            return None
        label = self.tabs.get_tab_label_text(self.tabs.get_nth_tab(page_number))
        return self.extract_number_from_name(label)
    
    def timers_object_changed(self, timers_object):
        pass
    
    def startup_object_changed(self, startup_object):
        pass
    
    def extract_number_from_name(self, name):
        """
        Extracts the number of a timer from its tab name.
        """
        if not name.lower().startswith("timer "):
            raise ValueError("Invalid timer name: " + name)
        name = name[len("timer "):]
        if ":" in name:
            name = name[:name.find(":")]
        return int(name)













"""
This module primarily contains Autobus's service discovery code.
"""

from abc import ABCMeta as ABC, abstractmethod as abstract
from socket import socket as Socket, AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_REUSEADDR, SO_BROADCAST, timeout as SocketTimeout, error as SocketError
from autobus2 import constants, net, exceptions
from threading import Thread
import json
from traceback import print_exc
import time
import random
from Queue import Queue


class Discoverer(object):
    """
    A class capable of discovering services available remotely and making a
    particular bus aware of those services.
    
    This class is abstract; one of its subclases must be used instead. Bus
    instances, by default, install a BroadcastDiscoverer when they are created.
    
    When a discoverer is started, it should start attempting to discover
    services, however it goes about doing that. When it discovers one, it
    should call bus.discovered(self, host, port, service_id, info_object).
    When one disappears, it should call bus.undiscovered, passing in the same
    parameters. If the info object changes, the discoverer should just call
    bus.discovered again with the new info object.
    """
    @abstract
    def startup(self, bus):
        """
        """
    
    @abstract
    def shutdown(self):
        """
        """


class BroadcastDiscoverer(object):
    """
    An implementation of Discoverer that listens for UDP broadcasts sent by
    other BroadcastPublisher instances on the network.
    """
    def __init__(self):
        self.running = True
        # Maps (host, port, service_id) to [info_object, last_alive_time]
        self.services = {}
        self.last_check_time = 0
    
    def startup(self, bus):
        if not self.running:
            raise Exception("BroadcastDiscoverers can only be used once")
        self.bus = bus
        self.sender, self.receiver = create_broadcast_sockets()
        Thread(name="autobus2.discovery.BroadcastDiscoverer.receive_loop", 
                target=self.receive_loop).start()
        Thread(name="autobus2.discovery.BroadcastDiscoverer.send_initial_requests", 
                target=self.send_initial_requests).start()
    
    def shutdown(self):
        with self.bus.lock:
            for k in self.services:
                self.bus.undiscover(self, *k)
            self.services.clear()
        self.running = False
        net.shutdown(self.receiver)
    
    def receive_loop(self):
        while True:
            message = None
            try:
                message, (host, port) = self.receiver.recvfrom(16384)
            except SocketError:
                pass
            if not self.running:
                net.shutdown(self.sender)
                return
            self.run_recurring()
            if not message:
                continue
            try:
                message = json.loads(message)
                if message["command"] == "add":
                    self.process_add(host, message["port"], message["service"], message["info"])
                elif message["command"] == "remove":
                    self.process_remove(host, message["port"], message["service"])
            except:
                print "Couldn't read message"
                print_exc()
                continue
    
    def send_initial_requests(self):
        for delay in constants.query_initial_intervals:
            for i in range(int(delay*10.0)): # Sleep in 100ms blocks so that if the
                # discoverer is shut down in the middle, we'll terminate fairly
                # quickly
                time.sleep(0.1)
                if not self.running:
                    return
            net.sendto(self.sender, json.dumps({"command": "query"}),
                    ("127.255.255.255", constants.broadcast_port))
            net.sendto(self.sender, json.dumps({"command": "query"}),
                    ("255.255.255.255", constants.broadcast_port))
    
    def run_recurring(self):
        check_interval = 12 # TODO: change this to more like 30, but make sure
        # it's greater than the Autobus connect timeout so that we don't get
        # two separate attempts to connect to the same service going on at the
        # same time
        expire_interval = 30 # TD: change this to more like 60
        current_time = time.time()
        if self.last_check_time + check_interval > current_time:
            return
        self.last_check_time = current_time
        # print "Checking discovered services..."
        with self.bus.lock:
            for k, v in self.services.iteritems():
                if v[1] + expire_interval < current_time: # Haven't received a
                    # message from this server in a while, so check to see if
                    # we can connect to it
                    Thread(name="autobus2.discovery.BroadcastDiscoverer.try_to_connect",
                            target=self.try_to_connect, args=(k, v)).start()
    
    def try_to_connect(self, k, v):
        host, port, service_id = k
        # print "Discovery timeout, attempting to connect to " + str(k)
        failed = False
        try:
            with self.bus.connect(*k) as connection:
                connection.wait_for_connect(10)
        except (exceptions.ConnectionException, exceptions.TimeoutException) as e:
            # print "Service could not be connected to, removing..."
            failed = True
        except:
            print "Unexpected failure while attempting to connect to service"
            print_exc()
            failed = True
        if failed:
            self.process_remove(*k)
                       
    def process_add(self, host, port, service_id, info):
        spec = (host, port, service_id)
        with self.bus.lock:
            if spec in self.services: # Spec is already there, so all we need
                # to do is check and make sure that the info object hasn't
                # changed, and then update the time that we last received
                # something from this service
                if self.services[spec][0] != info: # Info object changed
                    self.services[spec][0] = info
                    self.bus.discover(self, host, port, service_id, info)
                # Update timestamp
                self.services[spec][1] = time.time()
            else: # Spec isn't there, so we need to add it
                self.services[spec] = [info, time.time()]
                self.bus.discover(self, host, port, service_id, info)
    
    def process_remove(self, host, port, service_id):
        spec = (host, port, service_id)
        with self.bus.lock:
            if spec in self.services:
                del self.services[spec]
                self.bus.undiscover(self, host, port, service_id)


class Publisher(object):
    """
    A class capable of publishing information about services available to a
    particular Bus instance.
    
    This class is abstract; one of its subclasses must be used instead. Bus
    instances, by default, install a BroadcastPublisher when they are created.
    """
    @abstract
    def startup(self, bus):
        """
        Called by Bus instances when a publisher is installed onto them. The
        only argument is the bus onto which this publisher was installed.
        """
    
    @abstract
    def shutdown(self):
        """
        Called by Bus instances when a publisher is about to be removed. The
        bus will first call the publisher's remove function for every service
        it had previously added before calling shutdown.
        """
    
    @abstract
    def add(self, service):
        """
        Called to let this publisher know that a new service is available.
        """
    
    @abstract
    def remove(self, service):
        """
        Called to let this publisher know that a service is no longer available.
        """


class BroadcastPublisher(Publisher):
    """
    An implementation of Publisher that uses UDP broadcasts to let others know
    about available services. BroadcastDiscoverer is the matching discoverer
    that knows how to receive these broadcasts.
    """
    def __init__(self):
        self.running = True
        self.services = {}
        self.next_time = 0
    
    def startup(self, bus):
        if not self.running:
            raise Exception("BroadcastPublishers can't be re-used.")
        self.bus = bus
        self.sender, self.receiver = create_broadcast_sockets()
        Thread(name="autobus2.discovery.BroadcastPublisher.receive_loop", target=self.receive_loop).start()
    
    def shutdown(self):
        # print "Shutting down"
        self.running = False
        net.shutdown(self.receiver)
    
    def run_recurring(self):
        if self.next_time < time.time():
            self.next_time = time.time() + (constants.broadcast_interval
                + (random.random() * constants.broadcast_random))
            self.broadcast_services()
            
    def receive_loop(self):
        while True:
            message = None
            try:
                message = self.receiver.recvfrom(16384)[0]
            except SocketError:
                pass
            if not self.running:
                net.shutdown(self.sender)
                return
            self.run_recurring()
            if not message:
                continue
            try:
                message = json.loads(message)
                if message["command"] == "query":
                    self.broadcast_services()
            except:
                print "Couldn't read message in BroadcastPublisher"
                print_exc()
                continue
    
    def respond_to_query(self):
        time.sleep(random.random() * constants.query_response_random)
        self.broadcast_services()
    
    def broadcast_services(self):
        with self.bus.lock:
            for service in self.services.values():
                self.send_add(service)
    
    def send_add(self, service):
        broadcast = {"command": "add", "port": self.bus.port,
                    "service": service.id, "info": service.info}
        broadcast = json.dumps(broadcast)
        net.sendto(self.sender, broadcast, 
                ("127.255.255.255", constants.broadcast_port))
        net.sendto(self.sender, broadcast, 
                ("255.255.255.255", constants.broadcast_port))
    
    def send_remove(self, service):
        broadcast = {"command": "remove", "port": self.bus.port,
                    "service": service.id}
        broadcast = json.dumps(broadcast)
        # Send remove messages in the opposite order of discover messages
        net.sendto(self.sender, broadcast, 
                ("255.255.255.255", constants.broadcast_port))
        net.sendto(self.sender, broadcast, 
                ("127.255.255.255", constants.broadcast_port))

    def add(self, service):
        with self.bus.lock:
            # print "Service added: " + service.id
            self.services[service.id] = service
            self.send_add(service)
        
    
    def remove(self, service):
        with self.bus.lock:
            # print "Service removed: " + service.id
            del self.services[service.id]
            self.send_remove(service)


def create_broadcast_sockets():
    sender = Socket(AF_INET, SOCK_DGRAM)
    sender.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
    sender.setsockopt(SOL_SOCKET, SO_BROADCAST, True)
    sender.bind(("", 0))
    receiver = Socket(AF_INET, SOCK_DGRAM)
    receiver.settimeout(constants.broadcast_receiver_timeout)
    receiver.setsockopt(SOL_SOCKET, SO_REUSEADDR, True)
    receiver.setsockopt(SOL_SOCKET, SO_BROADCAST, True)
    receiver.bind(("", constants.broadcast_port))
    return sender, receiver



































from afn.utils import singleton
from collections import namedtuple

ANY = singleton.Singleton("autobus2.filter.ANY")
NOT_PRESENT = singleton.Singleton("autobus2.filter.NOT_PRESENT")

def filter_matches(info, filter):
    if filter is None:
        return True
    for k, v in filter.items():
        if v is ANY:
            if k not in info:
                return False
        elif v is NOT_PRESENT:
            if k in info:
                return False
        elif isinstance(v, NotEqualTo):
            if info.get(k) == v.value:
                return False
        else:
            if info.get(k) != v:
                return False
    return True


NotEqualTo = namedtuple("NotEqualTo", ["value"])



"""
This module provides a simple wiki backed by a git repository.

It's a really simple wiki for now. It doesn't even require a password to log
in; you simply enter your name and email address as you want them to show up in
the commit log, or leave them blank to use the name and email address
configured in the repository or on your machine.
"""



from BeautifulSoup import BeautifulSoup
from urllib import urlopen
import re
from collections import namedtuple
from traceback import print_exc
import json


ClassSeating = namedtuple("ClassSeating", ["id", "department", "number", "section", "title", "max", "current", "available"])


def extract_text(node):
    return " ".join([t for t in node.recursiveChildGenerator() if isinstance(t, basestring)])


def read_page(url):
    return urlopen(url).read()


def get_website(term=1124):
    """
    Gets the URL for the main campus schedule website using the specified term.
    The default, 1124, is Spring 2012; this value can be obtained by going to
    http://www.utah.edu/students/catalog.php, clicking on the appropriate link
    on the right, and examining the term= query parameter in the URL.
    """
    return "http://www.acs.utah.edu/uofu/stu/scheduling/?cmd=index&classtype=g&term=" + str(term)


def yield_departments(website=get_website(), progress=True):
    """
    A generator that yields tuples, each one of the format (id, name) where id
    is the identifier of a department and name is the name of the department.
    If a name is not presenet for a particular department, the empty string
    will be used for its name.
    """
    if progress:
        print "Downloading the department list..."
    page = BeautifulSoup(read_page(website))
    regex = re.compile(".*&dept=([A-Z]+)&.*")
    for a in page.findAll("a", href=regex):
        name = regex.match(a["href"]).group(1)
        desc_td = a.findParent("td").findNextSibling("td")
        if desc_td:
            description = "".join(desc_td.findChild("font").contents)
        else:
            description = ""
        yield name, description


def yield_class_seatings(department, term=1124, progress=True):
    url = "http://www.acs.utah.edu/uofu/stu/scheduling/crse-info?term=%s&subj=%s"
    url = url % (term, department)
    if progress:
        print "Downloading class seating information for department %s..." % department
    page = BeautifulSoup(read_page(url))
    trs = page.findAll("tr")
    for tr in trs:
        if tr.find("td") is None or tr.find("td").get("bgcolor", None) != "#eeeeee":
            continue
        tds = tr.findAll("td")
        t = [extract_text(td).strip() for td in tds]
        yield ClassSeating(t[0], t[1], t[2], t[3], t[4], int(t[5]), int(t[6]), int(t[7]))


def yield_all_available_class_seatings(term=1124, progress=True):
    for department_name, department_desc in yield_departments(get_website(term), progress):
        for seating in yield_class_seatings(department_name, term, progress):
            try:
                if seating.available > 0:
                    yield seating
            except:
                print_exc()


def write_all_seatings_to_file(filename):
    with open(filename, "w") as f:
        for department_name, department_desc in yield_departments():
            for seating in yield_class_seatings(department_name):
                f.write(json.dumps(seating._asdict()) + "\n")
    


































from threading import Thread, RLock
from Queue import Queue
from libautobus import read_fully, discard_args, create_message
from libautobus import InputThread, OutputThread, get_next_id
from libautobus import COMMAND, RESPONSE, NOTIFICATION
from libautobus import DEFAULT_PORT, NoSuchInterfaceException
from libautobus import NoSuchFunctionException, encode_object, decode_object
from libautobus import get_function_doc
from libautobus.message_types import *
from struct import pack, unpack
from json import dumps, loads
from traceback import print_exc
from socket import socket as Socket, SHUT_RDWR, SOL_SOCKET, SO_REUSEADDR
from functools import partial
import re
import sys
import inspect
import os
from autobus import processors, special_interface, special_objects

print """\
Autobus, a message bus that's kind of a cross of D-Bus and RPC, written
by Alexander Boyd (a.k.a. javawizard or jcp)
"""

autobus_interface = None
connection_map = {} # connection ids to connection objects
interface_map = {} # interface names to interfaces
listener_map = {} # Maps tuples of interface names and function names to lists
# of connection ids representing connections that should be notified when the
# specified listener is fired
watch_map = {} # Same as listener_map, but represents connections that should
# receive an object's value when it changes
event_queue = Queue() # A queue of tuples, each of which consists of a
# connection id and either a protobuf Message object or None to indicate that
# the connecton was lost and so should be shut down
processed_message_count = 0


def register_event_listener(interface_name, event_name, connection_id):
    """
    Adds a listener on the specified event. This only modifies the global listener
    map; this does not modify the connection's list of listeners., and this does
    not cause any RegisterListenerResponse to be sent back to the client.
    """
    event_spec = interface_name, event_name
    if event_spec not in listener_map:
        listener_map[event_spec] = []
    listener_map[event_spec].append(connection_id)

def deregister_event_listener(interface_name, event_name, connection_id):
    """
    Same as register_event_listener, but deregisters a listener previously added.
    """
    event_spec = interface_name, event_name
    listener_map[event_spec].remove(connection_id)
    if len(listener_map[event_spec]) == 0:
        del listener_map[event_spec]

def register_object_watch(interface_name, object_name, connection_id):
    """
    Adds a watch on the specified object. This only modifies the global watch
    map; this does not modify the connection's list of watches., and this does
    not cause any WatchObjectResponse to be sent back to the client.
    """
    object_spec = interface_name, object_name
    if object_spec not in watch_map:
        watch_map[object_spec] = []
    watch_map[object_spec].append(connection_id)

def deregister_object_watch(interface_name, object_name, connection_id):
    """
    Same as register_object_watch, but deregisters a watch previously added.
    """
    object_spec = interface_name, object_name
    watch_map[object_spec].remove(connection_id)
    if len(watch_map[object_spec]) == 0:
        del watch_map[object_spec]

def process_event_queue():
    """
    Starts processing the event queue. This function returns when (None, None)
    is pushed onto the event queue. This function generally should be started
    in a new thread.
    """
    global processed_message_count
    while True:
#        print "Waiting for events..."
        sender, message = event_queue.get(block=True)
        if sender is None: # pushed on at KeyboardInterrupt to shut down the event queue
#            print "Shutdown event received"
            break
        connection = connection_map.get(sender, None)
#        print ("Event " + str(message) + " received for sender " + str(sender)
#                + " and connection " + str(connection))
        processed_message_count += 1
        try:
            message(sender, connection)
        except:
            print "MAJOR ERROR in a protocol function handler:"
            print_exc()
            # os._exit(1)
    print "Shutting down connections..."
    connection_count = len(connection_map)
    for connection in connection_map.values():
        connection.shutdown_and_deregister()
    print str(connection_count) + " connection" + ("s" if connection_count
            != 1 else "") + " have been shut down."

class Function(object):
    """
    Represents a function registered on an interface.
    """
    def __init__(self, sender, name, doc, special_target=None):
        self.name = name
        self.sender = sender
        self.doc = doc
        self.special_target = special_target
        self.special = special_target != None
        if self.special:
            self.doc = inspect.getdoc(special_target)
            if self.doc is None:
                self.doc = ""
    
    def invoke_special(self, message, connection):
        args = message["arguments"]
        args = [decode_object(arg) for arg in args]
        try:
            return_value = self.special_target(*args)
        except Exception as e:
            print_exc()
            return_value = e
        result = encode_object(return_value)
        response = create_message(CallFunctionResponse,
                message, return_value=result)
        connection.send(response)

class Event(object):
    def __init__(self, sender, name, doc):
        self.name = name
        self.sender = sender
        self.doc = doc
        # TODO: allow special events
    
    def notify(self, arguments):
        """
        Sends a message to all connections listening for this event containing
        the specified arguments, which should be passed into this function as
        a (possibly empty) list or tuple.
        """
        listener_spec = (self.interface.name, self.name)
        if listener_spec in listener_map:
            message = create_message(FireEventCommand,
                    NOTIFICATION, interface_name=self.interface.name,
                    event_name=self.name, arguments=arguments)
            encoded_message = dumps(message)
            for connection_id in listener_map[listener_spec]:
                connection = connection_map[connection_id]
                connection.send(encoded_message)

    
class Object(object):
    def __init__(self, sender, name, doc, value):
        """
        Initializes this object. sender is the id of the connection that
        created this object. Name and doc are the object's name and documentation,
        respectively. value is the object's initial value.
        """
        self.name = name
        self.interface = None # Interface sets this when an object is registered
        # to an instance of it, so we don't have to
        self.sender = sender
        self.doc = doc
        self._value = value # This is an instance of protobuf.Instance
    
    def set_and_notify(self, new_value):
        """
        Sets this object's value to the specified value and sends a message to
        all connections watching the object with the new value. 
        """
        self._value = new_value
        self.notify()
    
    def notify(self):
        """
        Sends a message to all connections watching this object containing the
        object's current value. This only calls self.get() once, so it's not
        generally processor-intensive even when lots of connections are
        listening for changes to this object.
        """
        watch_spec = (self.interface.name, self.name)
        if watch_spec in watch_map:
            current_value = self.get()
            message = create_message(SetObjectCommand,
                    NOTIFICATION, interface_name=self.interface.name,
                    object_name=self.name, value=current_value)
            encoded_message = dumps(message)
            for connection_id in watch_map[watch_spec]:
                connection = connection_map[connection_id]
                connection.send(encoded_message)
    
    def get(self):
        """
        Returns the current value of the object. This is used instead of
        self._value to allow for subclasses to provide custom values for the
        object. VirtualObject in particular makes use of this.
        
        The return value must be an instance of protobuf.Instance.
        """
        return self._value

class VirtualObject(Object):
    """
    A subclass of Object that allows the object to have a value computed
    whenever it is requested. This is for server-side use only, and is used to
    implement most objects on the Autobus interface.
    """
    def __init__(self, name, doc, accessor):
        """
        Initializes this object. Name, and doc are the same as on
        the Object constructor. Accessor is a function that takes no
        arguments and returns the object's current value as a normal Python
        object. The resulting object will be converted to an encoded instance
        before being returned from the object's get() function.
        """
        Object.__init__(self, 0, name, doc, None)
        self.accessor = accessor
    
    def set_and_notify(self, new_value):
        raise Exception("VirtualObjects cannot have their value set")
    
    def get(self):
        value = self.accessor()
        return encode_object(value)

class Interface(object):
    """
    Represents an interface registered by a particular connection. Interfaces
    have an id, a name, and a reference to the connection that registered the
    interface. They also have a dict mapping names of functions to the
    corresponding function objects, and they have a single Instance
    representing more information about the interface.
    """
    def __init__(self, name, connection, doc):
        """
        Creates a new interface object. This does not register it with all of
        the global dicts nor the connection's dicts; register() must be called
        to do that.
        """
        self.name = name
        self.connection = connection
        self.doc = doc
        self.function_map = {} # Maps function names to Function instances
        self.event_map = {} # Maps event names to Event instances
        self.object_map = {} # Maps object names to Object instances
        self.special = False
    
    def register(self):
        """
        Registers this interface with the global dicts and the connection's
        dicts. This must only be called on the event thread.
        """
        if self.name in interface_map:
            raise KeyError()
        interface_map[self.name] = self
        if not self.special:
            self.connection.interfaces.append(self)
        # Notifications
        if self is not autobus_interface:
            autobus_interface.notify_object("interface_count")
            autobus_interface.notify_object("interfaces")
            autobus_interface.notify_object("interface_items")
    
    def register_special(self):
        """
        Registers this interface as a special interface. This sets the special
        attribute to True and completes registration normally. Special
        interfaces are those registered by the server (or any other server-side
        functionality). Their connection is, instead of being a connection, an
        object with functions directly on it.
        """
        self.special = True
        self.register()
        self.doc = inspect.getdoc(self.connection)
        if self.doc is None:
            self.doc = ""
    
    def deregister(self):
        """
        The opposite of register(): deregisters this interface.
        """
        print "Deregistering interface with name " + self.name
        del interface_map[self.name]
        self.connection.interfaces.remove(self)
        for object_name in self.object_map:
            self.object_map[object_name].set_and_notify(encode_object(None))
        # Notifications
        if self is not autobus_interface:
            autobus_interface.notify_object("interface_count")
            autobus_interface.notify_object("interfaces")
            autobus_interface.notify_object("interface_items")
    
    def register_function(self, sender, function_name, doc, notify=True):
        if function_name in self.function_map:
            raise KeyError()
        print ("Registering function " + function_name + " from " + 
                str(sender) + " on " + self.name) 
        function = Function(sender, function_name, doc)
        self.function_map[function_name] = function
        if notify and (self is not autobus_interface):
            autobus_interface.notify_object("interface_items")
        return function
    
    def register_event(self, event):
        if event.name in self.event_map:
            raise KeyError(event.name)
        print ("Registering event " + event.name + " from " + 
                str(event.sender) + " on " + self.name)
        event.interface = self
        self.event_map[event.name] = event
        if self is not autobus_interface:
            autobus_interface.notify_object("interface_items")
    
    def create_and_register_event(self, sender, event_name, doc):
        event = Event(sender, event_name, doc)
        self.register_event(event)
        return event
    
    def register_object(self, object):
        if object.name in self.object_map:
            raise KeyError(object.name)
        print ("Registering object " + object.name + " from " + 
                str(object.sender) + " on " + self.name)
        object.interface = self
        self.object_map[object.name] = object
        if self is not autobus_interface:
            autobus_interface.notify_object("interface_items")
        object.notify()
    
    def create_and_register_object(self, sender, object_name, doc, value):
        object = Object(sender, object_name, doc, value)
        self.register_object(object)
        return object
    
    def fire_event(self, name, arguments):
        """
        Looks up the specified event, then calls its notify method with the
        specified arguments.
        """
        self.event_map[name].notify(arguments)
    
    def notify_object(self, name):
        """
        Looks up the specified object, then calls its notify method.
        """
        self.object_map[name].notify()
    
    def lookup_function(self, function_name):
        """
        Looks up the function with the specified id or name. If this interface
        is special, this function requires the name to be specified, and a
        special function object wrapping the function in question will be
        returned.
        """
        if self.special:
            if function_name.startswith("_"): # Don't allow functions whose
                # names start with underscores
                raise NoSuchFunctionException(function_name)
            try:
                return Function(None, function_name, None, getattr(self.connection, function_name))
            except AttributeError:
                raise NoSuchFunctionException(function_name)
        if function_name in self.function_map:
            return self.function_map[function_name]
        raise NoSuchFunctionException(function_name)
    
    def describe_functions(self):
        """
        Returns a list of maps in the same format that the list_functions
        function of the Autobus internal interface returns. In fact, that
        function delegates to this one once it's looked up the interface.
        """
        result = []
        if self.special:
            function_list = dir(self.connection)
            function_list = [function for function in function_list if
                    not function.startswith("_")]
            for function in function_list:
                result.append({"name": function, "special": True,
                        "doc": get_function_doc(self.connection,
                                function)})
        else:
            for function in self.function_map.values():
                result.append({"name": function.name,
                        "special": function.special, "doc": function.doc})
        return result
    
    def describe_events(self):
        return [{"name": event.name, "doc": event.doc} 
                for event in self.event_map.values()]
    
    def describe_objects(self):
        return [{"name": object.name, "doc": object.doc}
                for object in self.object_map.values()]

class Connection(object):
    """
    Represents a connection. Connections are created when someone connects to
    autobus. They are destroyed when autobus exits or the connection is closed.
    Connections have an id, a list of interfaces they have registered, a queue
    containing protobuf Message objects to be sent to the corresponding
    socket as soon as possible, and an input and output thread.
    """
    def __init__(self, socket, address):
        self.socket = socket
        self.socket_address = address
        self.interfaces = []
        self.listeners = [] # List of tuples, each of which contains an
        # interface name and an event name.
        self.watches = [] # List of tuples, each of which contains an
        # interface name and an object name.
        self.message_queue = Queue()
        self.id = get_next_id()
        self.pending_responses = {} # Maps message ids of commands sent to this
        # connection to tuples consisting of the id of a connection that the
        # response should be forwarded to and the message id that the response
        # should have
        self.input_thread = InputThread(socket, self.message_arrived,
                                        self.input_closed)
        self.output_thread = OutputThread(socket, self.read_next_message)
    
    def register(self):
        """
        Registers this connection with the global connection map. This must
        only be called on the event thread.
        """
        print "New connection " + str(self.id) + " from " + str(self.socket_address)
        connection_map[self.id] = self
    
    def message_arrived(self, message):
        try:
            function = find_function_for_message(message)
        except:
            print "Client sent an invalid command (they will be disconnected):"
            print_exc()
            raise
        event_queue.put((self.id, partial(function, message)), block=True)
    
    def input_closed(self):
        event_queue.put((self.id, discard_args(self.shutdown_and_deregister)),
                        block=True)
        self.message_queue.put(None, block=True)
    
    def read_next_message(self):
        return self.message_queue.get(block=True)
    
    def close(self):
        """
        Closes the underlying socket for this connection. This, in turn, causes
        the input and output threads to shut down. This can be called from any
        thread.
        """
        self.socket.close()
    
    def send(self, message):
        """
        Adds the specified message to this connection's message queue. It will
        be sent to the underlying socket as soon as possible by the
        connection's output thread. This can be called from any thread.
        
        If the message is None, this function returns without doing anything.
        This allows for code that constructs a message pair with
        create_message_pair and then sends the message without worrying about
        whether or not create_message_pair returned an empty message (which it
        would if it was used to "reply" to a notification instead of to a
        command).
        """
        if message is not None:
            if isinstance(message, dict) and message["message_type"] is None:
                return
            self.message_queue.put(message, block=True)
    
    def send_new(self, *args, **kwargs):
        """
        Exactly the same as
        self.send(libautobus.create_message(*args, **kwargs)).
        """
        self.send(create_message(*args, **kwargs))
    
    def send_error(self, in_reply_to=None, **kwargs):
        """
        Sends an error back to the client. If in_reply_to is itself a
        notification message, no message will be sent back. Otherwise, an
        ErrorResponse to the specified message will be sent. The error's
        fields will be initialized to have any additional keyword arugments
        specified. You'll typically do:
        
        send_error(some_message, text="A random problem happened")
        """
        if in_reply_to == None:
            in_reply_to = NOTIFICATION
        message = create_message(ErrorResponse, in_reply_to, **kwargs)
        self.send(message)
    
    def start(self):
        """
        Starts this connection's input and output threads.
        """
        self.input_thread.start()
        self.output_thread.start()
    
    def shutdown_and_deregister(self):
        """
        Shuts this connection down and deregisters it. This also pushes error
        messages onto the message queues of any connections that were waiting
        for a response from this connection. This must only be called from the
        event thread.
        """
        print "Connection " + str(self.id) + " from " + str(self.socket_address) + " shutting down"
        try:
            self.socket.shutdown(SHUT_RDWR)
        except:
            pass
        self.socket.close()
        if self.id in connection_map:
            del connection_map[self.id]
        for connection_id, message_id in self.pending_responses.values():
            if connection_id in connection_map:
                connection_map[connection_id].send(create_error_response(
                        message_id, "The connection this response was waiting "
                        "on closed unexpectedly"))
        for interface in self.interfaces:
            interface.deregister()
        for interface_name, event_name in self.listeners:
            deregister_event_listener(interface_name, event_name, self.id)
        for interface_name, object_name in self.watches:
            deregister_object_watch(interface_name, object_name, self.id)

def create_error_response(message_id, text):
    return create_message(ErrorResponse, RESPONSE, message_id=message_id, text=text)

def find_function_for_message(message):
    action = message["action"]
    function_name = "process_" + action
    if function_name not in vars(processors):
        raise Exception("Message is missing processor function "
                        + function_name)
    return vars(processors)[function_name]

def lookup_interface(interface_name, owner=None):
    """
    Looks up an interface with the specified name. If there is no such
    interface, a NoSuchInterfaceException will be raised.
    
    If owner is not None, it should be the id of a connection. Only an
    interface registered by this connection will be returned.
    """
    if interface_name not in interface_map:
        raise NoSuchInterfaceException(interface_name)
    else:
        interface = interface_map[interface_name]
        if owner is not None and (interface.special or 
                interface.connection.id != owner):
            raise NoSuchInterfaceException(interface_name,
                    info="You're not the owner of that particular interface.")
        return interface


def register_builtin_interface():
    global autobus_interface
    interface = Interface("autobus", special_interface.AutobusInterface(), None)
    autobus_interface = interface
    interface.register_special()
    for name in dir(special_objects):
        if name[0] != "_":
            function = getattr(special_objects, name)
            if inspect.isfunction(function):
                if function.__doc__ is None:
                    function_doc = ""
                else:
                    function_doc = inspect.getdoc(function)
                interface.register_object(VirtualObject(name,
                        function_doc, function))

# MAIN CODE

def main():
    register_builtin_interface()
    server = Socket()
    if len(sys.argv) > 1:
        server_port = int(sys.argv[1])
    else:
        server_port = DEFAULT_PORT
    print "Listening on port " + str(server_port)
    server.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    server.bind(("", server_port))
    server.listen(50)
    
    Thread(target=process_event_queue).start()
    
    print "\nAutobus has successfully started up."
    
    try:
        while True:
            socket, address = server.accept()
            connection = Connection(socket, address)
            event_queue.put((connection.id, discard_args(connection.register)), block=True)
            connection.start()
    except KeyboardInterrupt:
        print "KeyboardInterrupt received, shutting down"
        event_queue.put((None, None), block=True)
        print "Event queue has been notified to shut down"
    except:
        print "Unexpected exception occurred in the main loop, shutting down. Stack trace:"
        print_exc()
        event_queue.put((None, None), block=True)
        print "Event queue has been notified to shut down"
    server.close()

    




        






















#!/usr/bin/env python

# DO NOT EDIT THIS FILE. Edit setup.py.template instead, then
# run update-setup-script to update this file.

#!/usr/bin/env python

from distutils.core import setup

version = "0.1.2012.05.17.21.57" # !!!VERSION!!!

setup(
    name="afn",
    version=version,
    author="Alexander Boyd",
    author_email="alex@opengroove.org",
    description="A collection of useful tools and libraries",
    long_description=
"""
**AFN** is Alexander Boyd's "labs" project. It contains a collection of useful
tools and libraries that aren't big enough or popular enough to merit their own
projects. 
""",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=[
        # !!!PACKAGES!!!
        'traytimer',
        'uofu',
        'autobus2.landfill',
        'autobus2.proxy',
        'autobus2.examples',
        'autobus2',
        'autobus2.tools',
        'libautobus',
        'librtk.examples',
        'librtk.schema',
        'librtk',
        'afn.ucreditu',
        'afn.landfill',
        'afn.dnsd.examples',
        'afn.dnsd',
        'afn.backports',
        'afn',
        'afn.utils',
        'afn.utils.cad',
        'afn.tools',
        'afn.bf',
        'afntest',
        'activehomed',
        'autobus',
    ],
    package_dir={"": "src"}
)




from utils import rotations

KEY_NAMES = [["A"], ["A#", "Bb"], ["B", "Cb"], ["C", "B#"], ["C#", "Db"],
             ["D"], ["D#", "Eb"], ["E", "Fb"], ["F", "E#"], ["F#", "Gb"],
             ["G"], ["G#", "Ab"]]

SEVENTH_CHORDS = {"major": ["C", "E", "G", "B"],
                  "minor": ["C", "Eb", "G", "Bb"],
                  "dominant": ["C", "E", "G", "Bb"],
                  "diminished": ["C", "Eb", "F#", "A"],
                  "half-diminished": ["C", "Eb", "F#", "Bb"],
                  "minor major": ["C", "Eb", "G", "B"],
                  "augmented major": ["C", "E", "G#", "B"],
                  "augmented": ["C", "E", "G#", "Bb"]}

def key_to_index(key):
    for index, names in enumerate(KEY_NAMES):
        if key in names:
            return index
    raise ValueError("No such key " + key)

def normal_index(index):
    return index % 12

def index_to_key(index):
    return KEY_NAMES[normal_index(index)][0]
    
SEVENTH_CHORD_NUMBERS = dict([
        (name, [
                (key_to_index(key) - key_to_index("C")) % 12
                for key in note_list
                ]) for name, note_list in SEVENTH_CHORDS.items()])

# Map of key indexes to map of seventh chord names to lists of four values,
# each value being that respective note in the seventh, as an index into
# KEY_NAMES
ALL_SEVENTHS = {}

for key_index in range(len(KEY_NAMES)):
    key_map = {}
    for seventh_name, seventh_notes in SEVENTH_CHORD_NUMBERS.items():
        key_map[seventh_name] = [normal_index(key_index + note)
                                  for note in seventh_notes]
    ALL_SEVENTHS[key_index] = key_map

def search_sevenths(first, second, third, fourth):
    """
    Prints out a list of seventh chords and their respective keys that match
    the specified chord. Usually there will only be one result; the general
    exception is when the notes form a diminished seventh, since any particular
    set of notes form a valid diminished seventh in four different keys.
    """
    first, second, third, fourth = (
            key_to_index(first), key_to_index(second),
            key_to_index(third), key_to_index(fourth)
            )
    for key, seventh_map in ALL_SEVENTHS.items():
        for seventh_name, note_list in seventh_map.items():
            if [first, second, third, fourth] in rotations(note_list):
                print "Match: " + KEY_NAMES[key][0].ljust(3) + seventh_name

























from autobus2 import Bus, wait_for_interrupt

class HelloService(object):
    def hi(self, text="world"):
        print "Saying hi to " + str(text)
        return "Hi, " + str(text) + "! How are you?"


def main():
    with Bus() as bus:
        bus.create_service({"autobus.example": "hello_server"}, from_py_object=HelloService())
        wait_for_interrupt()



"""
This module contains classes and functions relating to publishing services.
"""

from Queue import Queue, Empty

from autobus2 import net, messaging, exceptions, common
import autobus2
from utils import no_exceptions
from traceback import print_exc
from threading import Thread
import functools
from afn.utils import Suppress
from afn.utils.concurrent import synchronized_on
from afn.utils.multimap import Multimap as _Multimap
import itertools
import copy

class RemoteConnection(object):
    """
    A connection from a remote client attached to a particular local service.
    This is for Autobus's use only; external code won't ever need to see this.
    
    The class that external code receives instances of when connecting to a
    service is Connection, not RemoteConnection. When a Connection connects,
    the other end of the socket, the one on the listening end, will be a
    RemoteConnection.
    """
    def __init__(self, bus, socket):
        self.bus = bus
        self.socket = socket
        self.queue = Queue()
        self.service = None
        self.watched_objects = set()
        self.listened_events = set()
        net.OutputThread(socket, self.queue.get).start()
        net.InputThread(socket, self.received).start()
    
    def received(self, message):
        # print "Received: " + str(message)
        if message is None: # Connection lost
            self.cleanup()
            return
        try:
            if self.service is None: 
                if message["_command"] != "bind":
                    raise Exception("First message must be bind")
                name = message["service"]
                with self.bus.lock:
                    service = self.bus.local_services.get(name)
                    if not service:
                        self.send_error(message, "The service you attempted to bind to does not exist.")
                        self.close()
                        return
                    self.service = service
                    # The response includes the service since specifying None
                    # as the service causes the introspection service to be
                    # bound to, and this lets the client know what the
                    # introspection service's service id currently is
                    self.send(messaging.create_response(message, service=name))
                    return
            processor = getattr(self, "process_" + message["_command"], None)
            if not processor:
                self.send_error(message, "That command does not exist.")
            processor(message)
        except:
            print_exc()
            with no_exceptions:
                self.send_error(message, "An exception was thrown while processing that command.")
    
    def send(self, message):
        # print "Sending " + str(message)
        if message:
            self.queue.put(message)
    
    def send_error(self, command_or_id, reason):
        self.send(messaging.create_error(command_or_id, reason))

    def close(self):
        self.queue.put(None)
        net.shutdown(self.socket)
    
    def cleanup(self):
        with self.bus.lock:
            self.queue.put(None)
            self.bus.bound_connections.remove(self)
            net.shutdown(self.socket)
            if self.service:
                for name in self.watched_objects:
                    self.service.unwatch_object(self, name)
                for name in self.listened_events:
                    self.service.unlisten_for_event(self, name)
    
    def process_call(self, message):
        function = self.service.functions.get(message["name"])
        if not function:
            self.send_error(message, "That function does not exist.")
            return
        args = message["args"]
        if function.mode == autobus2.SYNC:
            self.send(function.call(message, args))
        elif function.mode == autobus2.THREAD:
            Thread(name="autobus2.local.RemoteConnection-function-caller",
                    target=lambda: self.send(function.call(message, args))).start()
        else:
            self.send(messaging.create_response(message, result=None))
            Thread(name="autobus2.local.RemoteConnection-function-async-caller",
                    target=functools.partial(function.call, message, args)).start()
    
    def process_watch(self, message):
        with self.bus.lock:
            # print "watch: ", message
            name = message["name"]
            if name in self.watched_objects:
                self.send_error(message, "You're already watching that object.")
                return
            self.watched_objects.add(name)
            self.service.watch_object(self, name)
            if self.service.objects.get(name):
                object_value = self.service.objects[name].value
            else:
                object_value = None
            self.send(messaging.create_response(message, name=name, value=object_value))
    
    def process_listen(self, message):
        raise NotImplementedError
    
    def process_unwatch(self, message):
        raise NotImplementedError
    
    def process_unlisten(self, message):
        raise NotImplementedError
    
    def process_ping(self, message):
        self.send(messaging.create_response(message))


class LocalService(common.AutoClose):
    """
    A service created locally and published to remote clients. This is the
    class of objects returned from bus.create_service().
    
    The various create_* functions can be called on LocalService instances to
    create functions, events, and objects that are published remotely.
    
    The function activate() can be called for services created by
    bus.create_service(..., active=False). This causes the service to become
    active. Inactive services are hidden from the outside world; this allows
    all of the functions needed for a particular service to be set up before
    it's published to the network.
    """
    def __init__(self, bus, id, info, doc):
        self.id = id
        self.info = info
        self.bus = bus
        self.doc = doc
        self.active = False
        self.is_alive = True
        self.functions = {}
        self.events = {}
        self.objects = {}
        self.event_listeners = _Multimap() # Maps event names to
        # RemoteConnection instances listening for them
        self.object_watchers = _Multimap() # Ditto but for object watches
    
    @synchronized_on("bus.lock")
    def activate(self):
        if self.active: # Already active
            return
        if not self.is_alive:
            raise exceptions.ClosedException("This LocalService has been closed.")
        self.active = True
        self.bus._i_update(self.id)
        for publisher in self.bus.publishers:
            publisher.add(self)
    
    @synchronized_on("bus.lock")
    def deactivate(self):
        if not self.active: # Not active
            return
        if not self.is_alive:
            return
        self.active = False
        for publisher in self.bus.publishers:
            publisher.remove(self)
        self.bus._i_update(self.id)
    
    @synchronized_on("bus.lock")
    def close(self):
        self.is_alive = False
        self.deactivate()
        self.bus._close_service(self)
    
    def create_function(self, name, function, mode=None, doc=""):
        with self.bus.lock:
            if mode is None:
                mode = autobus2.THREAD
            function = LocalFunction(self, name, function, mode, doc)
            self.functions[name] = function
            self.bus._i_update(self.id)
    
    def create_event(self):
        raise NotImplementedError
    
    def create_object(self, name, value, doc=""):
        with self.bus.lock:
            object = LocalObject(self, name, value, doc)
            self.objects[name] = object
            object.notify_created()
            self.bus._i_update(self.id)
            return object
    
    @synchronized_on("bus.lock")
    def use_py_object(self, py_object):
        for name in dir(py_object):
            if name.startswith("_"):
                continue
            value = getattr(py_object, name)
            if not callable(value):
                continue
            self.create_function(name, value)
    
    def watch_object(self, connection, name):
        self.object_watchers.add(name, connection)
    
    def unwatch_object(self, connection, name):
        self.object_watchers.remove(name, connection)
    
    def _add_introspection(self):
        self.create_object("autobus.details",
                self._i_details_function, "Provides information about the "
                "various functions, objects, and such provided by this interface")
        # Creating this function will cause the above object to update, so we
        # don't need to manually update it.
        self.create_function("autobus.get_details", self._i_details_function,
                doc="Returns the current value of the autobus.details object.")
    
    @synchronized_on("bus.lock")
    def _i_details_function(self):
        details = {}   
        details["id"] = self.id         
        details["active"] = self.active
        details["doc"] = self.doc
        details["info"] = self.info
        functions = {}
        events = {}
        objects = {}
        details["functions"] = functions
        details["events"] = events
        details["objects"] = objects
        for name, function in self.functions.items():
            functions[name] = {"name": name, "doc": function.doc}
        for name, event in self.events.items():
            events[name] = {"name": name, "doc": event.doc}
        for name, object in self.objects.items():
            objects[name] = {"name": name, "doc": object.doc}
        return details


class LocalFunction(object):
    def __init__(self, service, name, function, mode, doc):
        if mode not in (autobus2.SYNC, autobus2.ASYNC, autobus2.THREAD):
            raise ValueError("Invalid mode: " + str(mode))
        self.service = service
        self.name = name
        self.function = function
        self.mode = mode
        self.doc = doc
    
    def call(self, message, args):
        """
        Calls this function and returns a response for the specified message,
        which should be a call command.
        """
        try:
            result = self.function(*args)
            return messaging.create_response(message, result=result)
        except Exception as e:
            return messaging.create_error(message, "Remote function threw an exception: %s: %s" % (type(e), e))


class LocalEvent(object):
    pass


class LocalObject(object):
    def __init__(self, service, name, value, doc):
        self._value = value
        net.ensure_jsonable(self.get_value())
        self.service = service
        self.name = name
        self.doc = doc
    
    @property
    def value(self):
        return self.get_value()
    
    @value.setter
    def value(self, new_value):
        self.set_value(new_value)
    
    def get_value(self):
        if callable(self._value):
            value = self._value()
        else:
            value = self._value
        return copy.copy(value) # TODO: do we need to copy the value when
        # returning it here? We might only need to copy it when originally
        # setting it...
    
    def set_value(self, value):
        net.ensure_jsonable(value() if callable(value) else value)
        if not callable(value):
            value = copy.copy(value) # Make a copy so that modifications to the
            # original object don't affect us
        with self.service.bus.lock:
            self._value = value
            new_value = self.get_value()
            for watcher in self.service.object_watchers.get(self.name, []):
                watcher.send(messaging.create_command("changed", True, name=self.name, value=new_value, event="changed"))
    
    def changed(self):
        """
        Causes this object to act as if its value had changed. This is really
        only useful when this object's value has been set to a function; this
        indicates that the return value of the function has changed.
        """
        with self.service.bus.lock:
            self.set_value(self._value)
    
    def notify_created(self):
        with self.service.bus.lock:
            for watcher in self.service.object_watchers.get(self.name, []):
                watcher.send(messaging.create_command("changed", True, name=self.name, value=self.get_value(), event="created"))
    
    def remove(self):
        with self.service.bus.lock:
            del self.service.objects[self.name]
            for watcher in self.service.object_watchers.get(self.name, []):
                watcher.send(messaging.create_command("changed", True, name=self.name, value=None, event="removed"))
            with Suppress(KeyError):
                del self.service.object_watchers[self.name]



















































from libautobus import AutobusConnection
from time import sleep

class ExampleInterface(object):
    """
    An example interface. It doesn't do much.
    
    Basically, all it allows you to do is say hello.
    """
    def say_hello(self, name):
        """
        Formats a "hello world"-style message greeting the specified person by
        name and returns it. The argument must be a string or an exception will
        be raised.
        """
        print "We're about to say hi to " + name
        return "Hello, " + name + ". How are you?"

server = AutobusConnection()
server.add_interface("example", ExampleInterface())
server.connect()


# That's it for the actual autobus code. The rest of the code is a hack to get
# around python not letting the application die with ctrl+c.
try:
    while True:
        sleep(1)
except KeyboardInterrupt:
    print "Shutting down"
    server.shutdown()




try:
    import argparse
except ImportError:
    from afn.backports import argparse
from functools import partial
from autobus2 import Bus, wait_for_interrupt
import autobus2
import time
import json

description = """
A command-line Autobus client.
"""

parser = argparse.ArgumentParser(prog="autosend2", description="!!!INFO!!!", add_help=False)

parser.add_argument("type", nargs="?")
parser.add_argument("name", nargs="?")
parser.add_argument("values", nargs="*", metavar="value")

modes = parser.add_argument_group("modes").add_mutually_exclusive_group()
add_mode = partial(modes.add_argument, action="store_const", dest="mode")
add_mode("-f", "--call", const="call",
        help="Function call mode")
add_mode("-b", "--broadcast", const="broadcast",
        help="Broadcast watch mode")
add_mode("-d", "--discovery", const="discovery",
        help="Discovery mode. In this mode, autosend2 prints a table of all "
        "services available on the network, and then prints a message whenever "
        "a service appears or disappears.")
add_mode("-l", "--list", const="list",
        help="List mode. In this mode, autosend2 prints out information about "
        "all matching services, including their documentation and a list of "
        "all functions, events, etc on the service and their documentation.")
add_mode("-?", "--help", const="help",
        help="Show this help message")

options = parser.add_argument_group("optional arguments")
options.add_argument("-k", "--filter-equal", action="append", dest="filter_equal",
        nargs=2, metavar=("key", "value"), default=[],
        help="The specified key must have the specified value.")
options.add_argument("-m", "--multiple", action="store_true", help=
        "Causes commands such as --call and --list to try to connect to as "
        "many services as possible. Those commands normally connect to the "
        "first service that they can; with --multiple, they will wait up to "
        "however long is specified by -t (which defaults to 2 seconds) and "
        "try to connect to as many services as they can during that time.")
options.add_argument("-t", "--time", action="store", type=int, default=2, help=
        "Specifies the amount of time that autosend2 should try to connect to "
        "services for when -m is used. This has no effect when -m is not used.")

def main():
    args = parser.parse_args()
    if args.mode is None and args.type is None:
        print "Use autosend2 --help for more information."
        return
    if args.mode == "help":
        print parser.format_help().replace("!!!INFO!!!", description.strip())
        return
    
    mode = args.mode
    if mode is None:
        # Type, at least, will be present, since the "Use autosend2 --help ..."
        # message will have stopped us if it wasn't. So basically we need to
        # check to see if name is present, and if it is, then we call the
        # specified function, and if it isn't, then we search for the
        # introspection service and print a list of all supported functions.
        if args.name is not None:
            mode = "call"
        else:
            mode = "list"
    
    info_filter = parse_info_filter(args)
    
    with Bus() as bus:
        if mode == "discovery":
            print discovery_mode_header
            bus.add_service_listener(discovery_mode_listener, info_filter=info_filter, initial=True)
            wait_for_interrupt()
            print 
            return
        if mode == "call":
            if not args.name:
                print "You need to specify the name of the function to call."
                return
            if args.multiple:
                with bus.get_service_proxy(info_filter, bind_function=
                        partial(call_mode_onbind, args), multiple=True) as proxy:
                    time.sleep(args.time)
            else:
                with bus.get_service_proxy(info_filter) as proxy:
                    proxy.wait_for_bind(timeout=args.time)
                    try:
                        call_mode_print_result(proxy[args.name](*parse_value_list(args.values)))
                    except Exception as e:
                        call_mode_print_result(e)
            return
        if mode == "list":
            with bus.get_service_proxy(info_filter, multiple=args.multiple) as proxy:
                if args.multiple:
                    time.sleep(args.time)
                else:
                    proxy.wait_for_bind(timeout=args.time)
                results = proxy["autobus.get_details"]()
                if not args.multiple:
                    results = {proxy.current_service_id: results}
                if len(results) == 0:
                    print "No matching services found."
                for i, (service_id, details) in enumerate(results.items()):
                    if i > 0:
                        print "#" * 70
                    
                    print "Service " + str(service_id) + ":"
                    if details["doc"]:
                        print "\n" + details["doc"]
                    print "=" * 70
                    functions = filter_hidden(args, details["functions"])
                    if len(functions) == 0:
                        print "No functions available on this service."
                    for j, function in enumerate(functions.values()):
                        if j > 0:
                            print "-" * 70
                        print "Function " + function["name"] + ":"
                        if function["doc"]:
                            print "\n" + function["doc"]
                    print "=" * 70
                    events = filter_hidden(args, details["events"])
                    if len(events) == 0:
                        print "No events available on this service."
                    for j, event in enumerate(events.values()):
                        if j > 0:
                            print "-" * 70
                        print "Event " + event["name"] + ":"
                        if event["doc"]:
                            print "\n" + event["doc"]
                    print "=" * 70
                    objects = filter_hidden(args, details["objects"])
                    if len(objects) == 0:
                        print "No objects available on this service."
                    for j, object in enumerate(objects.values()):
                        if j > 0:
                            print "-" * 70
                        print "Object " + object["name"] + ":"
                        if object["doc"]:
                            print "\n" + object["doc"]
                return
        print "Unsupported mode used: " + str(mode)


def parse_info_filter(args):
    filter = {}
    filter.update(dict(args.filter_equal))
    if args.type:
        filter["type"] = args.type
    return filter


discovery_mode_header = ("Time".ljust(26) + "Event".ljust(13) + "Service ID".ljust(42)
        + "Host".ljust(15) + "Port".ljust(7) + "Info")

def discovery_mode_listener(service_id, host, port, info, event):
    text = time.ctime(time.time()).ljust(26)
    if event is autobus2.DISCOVERED:
        text += "DISCOVERED".ljust(13)
    elif event is autobus2.UNDISCOVERED:
        text += "REMOVED".ljust(13)
    else:
        text += "CHANGED".ljust(13)
    text += service_id.ljust(42)
    text += host.ljust(15)
    text += str(port).ljust(7)
    short_info = dict(info)
    del short_info["host"]
    del short_info["port"]
    del short_info["service"]
    text += str(short_info)
    print text


def call_mode_print_result(result):
    if isinstance(result, Exception):
        print type(result).__name__ + ": " + str(result)
    else:
        print result

def call_mode_onbind(args, proxy, connection, info):
    connection[args.name](*parse_value_list(args.values), safe=True,
            callback=call_mode_print_result)


def parse_value_list(values):
    return [parse_value(v) for v in values]


def parse_value(value):
    if value.startswith("="):
        return value[1:]
    if value.startswith("@"):
        return json.loads(value)
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    return value


def filter_hidden(args, dictionary):
    new_dict = {}
    for k, v in dictionary.items():
        if not k.startswith("autobus."):
            new_dict[k] = v
    return new_dict




































from utils import print_exceptions
from librtk.various import tracker

"""
Identical to count, but the counter is shared among clients so that
clicking the button in one client causes all others to see the new value.

This example is not perfect; it has some race conditions because it doesn't
use locks where it should. I may fix this at some point, but this is
intended to be a proof-of-concept, not a perfect application.
"""

shared_count_state = 0
shared_count_clients = []

@tracker(shared_count_clients)
def start(connection):
    window = connection.Window(title="The Shared Counter")
    window.close_request.listen(connection.close)
    box = connection.VBox(window)
    label = connection.Label(box, text=str(shared_count_state))
    button = connection.Button(box, text="Click to increment the counter "
            "across all clients.")
    def clicked():
        global shared_count_state
        shared_count_state += 1
        for s_con in shared_count_clients[:]:
            with print_exceptions:
                s_con[0][0][0].text = str(shared_count_state)
    button.clicked.listen(clicked)







"""
Let's figure out some things about how this is going to work...

Functions are passed around as Python functions. Numbers are represented by
decimal.Decimal instances. Strings and booleans are represented by basestring
and bool instances. Lists and maps are represented by Python lists and dicts.
(Tuples don't function correctly at present. I hope to add support for them at
some point in the future, but I haven't yet decided how to go about this.) null
and void are represented by libfact. Null and None, respectively.

Functions are registered to a Interpreter individually. They can be either
normal or Fact-aware. By default, a function registered to an interpreter will
be registered as a normal function. The function has to be decorated with
@libfact.function in order to be a Fact-aware function. Fact-aware functions
are different in two ways: they can specify (via arguments to the decorator)
which arguments should be passed as values and which arguments should be passed
as closures, and they have an additional first argument, an instance of
libfact.Context which allows the function access to variables and other context
information. Closures are passed as callable objects that can be called with
a single optional argument, the context under which the closure should be run.
If the context is not specified, the context passed into the function into
which the closure was passed will be used.

Closures return what they evaluate to when run. The return value of a function
is the value that the function call will evaluate to in the Fact script. Since
void is represented in python by None, a function that does not return a value
will evaluate to void.

The dot operator only works on maps and lists. For these, it subscripts them.
If the result is None, it converts it to Null. Using the dot operator on Null
evaluates to Null, so a list of dot operations can be chained without fear
that one of them being null will cause an error.

Variables in a context can be accessed and written by subscripting the context.
Internally, they're stored in the context's vars attribute, which can be any
subscriptable object. The context itself should be used where possible, though,
as it translates some values (for example, it translates int, long, and float
to Decimal).

The parser's architecture is a bit different from the original Fact parser.
Instead of having one function that parses function references, there are two
functions: one parses closures and one parses function calls. The parser
returns a callable object that accepts one required argument, namely a context
in which it's supposed to be run. These are all subclasses of Entity.

The subclasses of Entity are Sequence, VarReference, Literal, and FunctionCall.
Each one has a function, resolve, that takes the context in which it should be
resolved. There's a class, Closure, that wraps this and tracks a
previously-existing context. Closure is callable and accepts one optional
argument, a context that will override the context it was created with.
Instances of Closure are passe to Fact-aware functions that request arguments
be passed as closures.

Sequences resolve themselves by resolving each of their arguments in turn.
They then call implicit_cat on their arguments. This function performs an implicit
concatenation on its values. VarReferences resolve themselves by evaluating
each of their dot components in turn and returning the result after this. If
there are no components, the empty string is returned. Literals simply return
a string representing the contents they were created with. FunctionCalls
resolve their first argument, then use it to ask the Interpreter they were
created against for the Function instance representing the named function.
They then call it, passing in the entities representing the function's
arguments. The Function class automatically wraps these in Closure instances
as needed, resolves those that are to be passed as values to the function, and
calls the function.

There are two main functions in charge of factoid parsing: parse_closure and
parse_function. The former parses Literals and VarReferences (and backslash
escapes) into a Sequence until it hits a | or a }, at which point it stops and
returns the Sequence. The latter expects its first input character to be a {.
It drops it, then calls parse_closure repeatedly. In between each invocation,
it checks to see what the next character is. If it's |, it continues with the
invocations. If it's }, it consumes the } and returns. Both of those functions
expect lists in reverse order (I.E. the last character in the list is the first
character in the text to parse and so on; this is because removing from the
end of a list in Python is usually several orders of magnitude more efficient
than removing from the beginning of a list). The parse function expects a
string. It converts it to a reversed list and parses it as a closure. It also
throws an exception if there's any input left after parsing, which would
indicate an extraneous | or } character.
"""

from decimal import Decimal

class NullType(object):
    """
    The type of the Fact Null object. Null and NullType are analogous to None
    and NoneType; None is used to represent void in Fact, so Null is used to
    represent null. Similarly to NoneType, this class should not be
    instantiated; the singleton instance Null should be used instead.
    """
    pass

Null = NullType()

class Interpreter(object):
    """
    A Fact interpreter. This is the main class that you'll use to interpret
    and run Fact scripts. At its simplest, this class can be used thus:
    
    >>> from libfact import Interpreter, Context
    >>> Interpreter().parse("One plus two is {+|1|2}.").resolve(Context())
    
    which would print "One plus two is 3.". Each individual interpreter
    maintains a map of functions available to scripts run on it. Functions
    can be registered using Interpreter.register_function.
    """
    def __init__(self, install_defaults=True):
        """
        Creates a new interpreter. If install_defaults is true, all of the
        functions returned by make_default_functions will be automatically
        registered to this interpreter.
        """
        self.functions = {}
        if install_defaults:
            for function in _default_functions:
                self.register_function(function)
    
    def register_function(self, *args):
        """
        register_function(name=None, function)
        
        Registers the specified function. If name is specified, it's used as
        the name for the function. Otherwise, the function's fact_name
        attribute is used as the name, or function.__name__ if the function
        does not have any such attribute.
        """
        if len(args) == 1:
            function = args[0]
            try:
                name = function.fact_name
            except AttributeError:
                name = function.__name__
            if name is None:
                raise Exception("Function does not have a name")
        else:
            name, function = args
        self.functions[name] = function
    
    def lookup_function(self, name, context):
        """
        Returns a Function wrapper around the specified function. This is used
        by the Fact interpreter and is generally not needed by outside code.
        """
        if name not in self.functions:
            raise Exception("No such function: " + str(name))
        return Function(self.functions[name], context)
    
    def parse(self, text):
        """
        Parses the specified Fact script into an instance of one of the
        subclasses of Entity. The resulting syntax tree is bound to this
        interpreter but is not otherwise state-bound; it can be resolved
        multiple times to run the same script multiple times without the
        overhead of reparsing it.
        
        The resolve(context) function can be called on the return value of
        this function, passing in an instance of Context. This will cause the
        script to be actually run; the return value from resolve() will be the
        value that the script evaluated to.
        """
        chars = list(text)
        chars.reverse()
        result = self.parse_closure(chars)
        if chars:
            raise Exception("Extra | or } in the script. Remember, | cannot "
                    "appear outside of a function call. You can escape it "
                    "with a backslash if you need an actual | in the text.")
        return result
    
    def parse_function(self, chars):
        """
        Parses the specified char stack, which should start with a function
        call. This is used by the Fact interpreter and is generally not needed
        by outside code.
        """
        if chars[-1] != "{":
            raise Exception("Functions must start with {")
        items = []
        while chars.pop() != "}":
            items.append(self.parse_closure(chars))
            if not chars:
                raise Exception("Extra { in the specified script")
        return FunctionCall(self, items)
    
    def parse_closure(self, chars):
        """
        Parses the specified char stack, which should start with a closure.
        This parses until the end of the current closure. This is used by the
        Fact interpreter and is generally not needed by outside code.
        """
        items = []
        while(chars):
            char = chars.pop()
            if char == "{":
                # Push the { back on
                chars.append("{")
                # ...and then parse it as a function
                items.append(self.parse_function(chars))
            elif char in "|}":
                # End of this argument
                chars.append(char) # Push back so that parse_function can tell
                # whether this was a | or a }
                break
            elif char == "%":
                # Var reference. We'll get all the chars up to the next %
                var = ""
                while chars[-1] != "%":
                    var += chars.pop()
                chars.pop() # Get rid of the trailing %
                items.append(VarReference(var.split(".")))
            else:
                if char == "\\": # Process backslash escapes
                    char = process_escape(chars)
                # Check to see if the last item is a literal. If so, we'll just
                # add this char to it to avoid creating one Literal per char
                if items and isinstance(items[-1], Literal):
                    items[-1].append(char)
                else: # Last item is not a literal, so we'll create a new one
                    items.append(Literal(char))
        if len(items) == 0:
            return Void()
        if len(items) == 1:
            return items[0]
        return Sequence(items)

class Context(object):
    def __init__(self, vars=None):
        """
        Creates a new context. The specified object, which should be
        subscriptable, will be used as the local variable store. If it's not
        specified, a newly-created dictionary will be used.
        """
        if vars is None:
            vars = {}
        self.vars = vars
        
    def __getitem__(self, item):
        """
        Gets the specified local variable from this context, translating it
        with foreign_get if need be before returning it.
        """
        return foreign_get(self.vars, item)
    
    def __setitem__(self, item, value):
        """
        Sets the specified local variable on this context. No translation is
        performed by this function.
        """
        self.vars[item] = value
    
    def varcopy(self, vars):
        """
        Creates a new context that mirrors (and delegates to) this one but uses
        the variables in the specified subscriptable object instead. 
        """
    
    def create_static_container(self, vars):
        """
        Creates a new context that mirrors (and delegates to) this one but uses
        a new StaticVarContainer to hold its variables. The container will
        point back to this context's variables and will use the specified
        mapping object as the static variable set.
        """
        return Context(StaticVarContainer(self.vars, vars))

class Function(object):
    """
    A wrapper around a Python function representing a function available to
    the Fact interpreter. This class is what actually runs a function.
    """
    def __init__(self, function, context):
        self.function = function
        self.context = context
    
    def run(self, entities):
        try:
            preserve_list = self.function.fact_preserve
            preserve_other = self.function.fact_preserve_other
            use_context = self.function.fact_context
        except AttributeError:
            preserve_list = []
            preserve_other = False
            use_context = False
        args = []
        for index, entity in enumerate(entities):
            if index >= len(preserve_list):
                preserve = preserve_other
            else:
                preserve = preserve_list[index]
            if preserve:
                entity = Closure(entity, self.context)
            else:
                entity = entity.resolve(self.context)
            args.append(entity)
        if use_context:
            args[0:0] = [self.context]
        return self.function(*args)

class Closure(object):
    """
    A closure object. Instances of this class are passed into functions
    registered with Interpreter.register_function for arguments that are to be
    preserved as closures. Calling an instance of this class causes the
    specified closure to be evaluated.
    """
    def __init__(self, entity, context):
        self.entity = entity
        self.context = context
    
    def __call__(self, context=None):
        """
        Evaluates this closure and returns the result. If context is None (the
        default), the context used to invoke the function that the instance is
        being passed into will be used.
        """
        if context is None:
            context = self.context
        return self.entity.resolve(context)

class Entity(object):
    """
    A Fact AST node. Calling resolve actually runs the node.
    """
    def resolve(self, context):
        raise Exception("resolve not implemented by an entity")
    
    def __repr__(self):
        return self.__str__()

class Void(Entity):
    """
    An entity that always resolves to None (which represents the Fact value
    void). This is used to represent closures that have no contents.
    """
    def resolve(self, context):
        return None
    
    def __str__(self):
        return "<Void>"

class Literal(Entity):
    """
    An entity that resolves to a piece of literal text.
    """
    def __init__(self, text):
        """
        Creates a Literal that will resolve to the specified text.
        """
        self.text = text
    
    def resolve(self, context):
        return self.text
    
    def append(self, text):
        """
        Appends some text to the text that this literal will resolve to.
        """
        self.text += text
    
    def __str__(self):
        return "<Literal: " + repr(self.text) + ">"

class VarReference(Entity):
    """
    An entity that resolves to the value of a particular variable. This
    represents %foo.bar.baz% constructs.
    """
    def __init__(self, strings):
        """
        Creates a new VarReference with the specified identifiers, which
        should be a list. The construct %foo.bar.baz% is parsed as
        VarReference(["foo", "bar", "baz"]).
        """
        self.strings = strings
    
    def resolve(self, context):
        if len(self.strings) == 0:
            return ""
        value = context[self.strings[0]]
        for specifier in self.strings[1:]:
            value = foreign_get(value, specifier)
        return value
    
    def __str__(self):
        return "<VarReference: " + ".".join(self.strings) + ">"

class FunctionCall(Entity):
    """
    An entity that resolves its first argument, which must result in a
    string, then looks up the function with the specified name and calls it
    with the rest of the arguments being passed in.
    """
    def __init__(self, interpreter, items):
        """
        Creates a new FunctionCall under the specified interpreter (which is
        used to look up the name of the function to run at invocation time)
        and with the specified items. {foo|bar|baz} is parsed as
        FunctionCall(interpreter, Literal("foo"), Literal("bar"),
        Literal("baz")).
        """
        self.interpreter = interpreter
        self.items = items
    
    def resolve(self, context):
        function_name = self.items[0].resolve(context)
        if not isinstance(function_name, basestring):
            raise Exception("Function name has to be a string; it was "
                    + str(function_name))
        function = self.interpreter.lookup_function(function_name, context)
        return function.run(self.items[1:])
    
    def __str__(self):
        return "<FunctionCall: " + str(self.items) + ">"

class Sequence(Entity):
    """
    An entity representing a sequence of entities occurring one after another.
    Resolving a sequence causes its items to be resolved in order, and the
    implicit concatenation (as performed by the implicit_cat function)
    returned.
    """
    def __init__(self, items):
        """
        Creates a new sequence with the specified items. "Hello %name%!", for
        example, would be parsed as Sequence([Literal("Hello "),
        VarReference(["name"]), Literal("!")]).
        """
        self.items = items
    
    def resolve(self, context):
        return implicit_cat([i.resolve(context) for i in self.items])
    
    def __str__(self):
        return "<Sequence: " + str(self.items) + ">"
    
class StaticVarContainer(object):
    """
    A subscriptable object that additionally provides a set of defaults, or
    static variables. Upon attempting to access a particular variable, the
    static variable container is checked, and if it contains the variable, the
    value from the static container is returned. Otherwise the normal container
    is checked. Upon attempting to set a particular variable, the static
    variable container is checked, and if it contains the variable, an
    exception indicating the variable is read-only is thrown. Otherwise, the
    variable is set into the normal container.
    """
    def __init__(self, container, static_vars):
        """
        Creates a new StaticVarContainer with the specified normal container
        and static container.
        """
        self.container = container
        self.special_vars = static_vars
    
    def __getitem__(self, item):
        if item in self.special_vars:
            return self.special_vars[item]
        return self.container[item]
    
    def __setitem__(self, item, value):
        if item in self.special_vars:
            raise Exception("Assignment to a static variable is not allowed.")
        self.container[item] = value
    
    def __contains__(self, item):
        return item in self.special_vars or item in self.container

def implicit_cat(values):
    """
    Computes the implicit concatenation of the specified values, which should
    be some sort of sequence. First, a new list is created containing all items
    in the specified list that are not None. If there's no elements in the
    resulting list, None is returned. If there's one element, that element is
    returned. If there are more, the values are converted to strings with str()
    and the resulting strings concatenated together (with no characters in
    between each string) and the resulting concatenation returned.
    """
    values = [value for value in values if value is not None]
    if len(values) == 0:
        return None
    if len(values) == 1:
        return values[0]
    return "".join(str(value) for value in values)

def foreign_get(object, item):
    """
    Gets the specified item from the specified object. If the object is None
    or Null, None is returned. If the object is an instance of list or tuple,
    item is converted to an integer first. Then a value is obtained by calling
    object[item], and the result passed into foreign_translate and then
    returned. If object[item] throws a KeyError or IndexError, None will be
    returned instead. This is used to implement %x.y% in the interpreter.
    """
    if object is None or object is Null:
        return None
    try:
        if isinstance(object, (list, tuple)):
            item = int(item)
        return foreign_translate(object[item])
    except (KeyError, IndexError):
        return None

def foreign_translate(object):
    """
    Translates a foreign Python object into a Fact-compatible object. If the
    argument is None, Null is returned. If the argument is an instance of int,
    long, or float, Decimal(object) is returned. Otherwise, the object is
    returned as-is.
    """
    if object == None:
        return Null
    if isinstance(object, (int, long, float)):
        return Decimal(object)
    return object

def process_escape(chars):
    char = chars.pop()
    if char == "n":
        return "\n"
    if char == "r":
        return "\r"
    if char == "t":
        return "\t"
    if char == "x":
        return chr(int(chars.pop() + chars.pop(), 16))
    return char

def function(*args, **kwargs):
    """
    Decorator that can be applied to functions that are to be registered with
    Interpreter.register_function. The decorator modifies the function
    in-place instead of creating a new one. Each positional argument
    corresponds to a possible argument to the function. If it's False, as is
    the default (both when extra arguments are provided and this decorator is
    not present), that particular argument will be resolved and passed in as
    a value. If it's True, the arguments will be preserved as a closure that
    will then be passed in. The closure will be a callable object accepting
    one additional optional parameter, which is the context under which the
    closure should be evaluated. If it's not specified, the context in which
    this function was called will be used.
    
    Three keyword arguments can additionally be present: other, name, and
    context. Other specifies whether or not any additional variables provided
    to the function will be preserved as closures. Name specifies the name of
    a function, which will default to the Python name of the function if
    unspecified. Context is a boolean indicating whether the context under
    which the function is being evaluated will be passed as the first
    argument (and the first Fact argument to the function passed as the second
    argument, etc). It defaults to True if this annotation is present but
    context is unspecified and False if this annotation is not present at all.
    """
    other = kwargs.get("other", None)
    name = kwargs.get("name", None)
    context = kwargs.get("context", True)
    def decorator(function):
        function.fact_preserve = args
        function.fact_preserve_other = other
        if name is not None:
            function.fact_name = name
        function.fact_context = context
        return function
    return decorator

def make_default_functions():
    @function(name="+", context=False)
    def _add(*args):
        return reduce(lambda x, y: x + y, (Decimal(i) for i in args))

    @function(name="-", context=False)
    def _subtract(*args):
        return reduce(lambda x, y: x - y, (Decimal(i) for i in args))
    
    @function(name="*", context=False)
    def _multiply(*args):
        return reduce(lambda x, y: x * y, (Decimal(i) for i in args))
    
    @function(name="/", context=False)
    def _divide(*args):
        return reduce(lambda x, y: x / y, (Decimal(i) for i in args))

    @function(name="//", context=False)
    def _floordiv(*args):
        return reduce(lambda x, y: x // y, (Decimal(i) for i in args))
    
    @function(name="2pad0", context=False)
    def _twopad0(value):
        return str(value).rjust(2, "0")
    
    def identity(value):
        return value
    
    @function(False, context=False, other=True)
    def switch(value, *args):
        for i in range(0, len(args) - 1, 2): # Iterate over every pair of
            # elements, excluding the odd last element if there is one
            if args[i]() == value:
                return args[i + 1]()
        if (len(args) % 2) == 1:
            return args[-1]()
        else:
            return None
    
    def extrange(value, div, mod):
        """
        Syntax: {extrange|<value>|<div>|<mod>} -- Divides <value> by <div>,
        then evaluates to that result modulo <mod>. This is useful for
        programs such as a timer client; assuming <x> holds a possibly-large
        number of seconds, {extrange|<x>|1|60} would extract the number of
        seconds for display on a digital clock, {extrange|<x>|60|60} would
        extract the number of minutes, and {extrange|<x>|3600|60} would extract
        the hours.
        """
        return (Decimal(value) // Decimal(div)) % Decimal(mod)
    
    @function(False, False, False, True, False, name="for")
    def for_function(context, start, stop, var, action, delimiter):
        varmap = {}
        new_context = context.create_static_container(varmap)
        result = []
        for i in range(int(start), int(stop) + 1):
            varmap[var] = i
            result.append(action(new_context))
            result.append(delimiter)
        if result:
            result = result[:-1]
        return implicit_cat(result)
    
    @function(name="list", context=False, other=False)
    def _list(*args):
        return list(args)
    
    @function(False, False, name="set")
    def _set(context, name, value):
        context[name] = value
    
    @function(False)
    def get(context, name):
        return context[name]
    
    return [_v for _k, _v in locals().copy().items()
            if not (_k.startswith("__") or _k.startswith("_["))]
    # The cukoobirdie syntax with checking for _[ is that for some totally
    # random reason CPython seems to leak the list comprehension currently in
    # progress into locals() as _[1] or some such. Does anyone know what the
    # heck is up with that?

_default_functions = make_default_functions()





































"""
This module contains classes and functions relating to connecting to and using
remote services.
"""

from Queue import Queue, Empty
from autobus2 import net, messaging, exceptions, common
from utils import Suppress, print_exceptions, Consume
from threading import Thread, RLock, Condition
import json
import autobus2
from traceback import print_exc
import time
from concurrent import synchronized_on
from socket import socket as Socket, error as SocketError, timeout as SocketTimeout
from afn.utils import full_name
from afn.utils.multimap import Multimap


class Connection(common.AutoClose):
    """
    A connection to a remote service. This class allows for calling functions
    on the remote service, listening for events, and watching objects.
    
    Instances of this class should not be created directly; instead, a new Bus
    object should be created, and its connect function called. The connect
    function will then return an instance of this class.
    """
    def __init__(self, bus, host, port, service_id, timeout, open_listener, close_listener, fail_listener, lock=None):
        """
        Creates a new connection, given the specified parent bus and socket.
        This constructor sets up everything and then sends an initial message
        to the remote socket indicating what service is to be connected to.
        """
        self.context_enters = 0
        self.bus = bus
        self.socket = None
        self.would_be_socket = None
        self.host = host
        self.port = port
        self.service_id = service_id
        self.queue = None
        self.query_map = {}
        if lock is None:
            lock = RLock()
        self.lock = lock
        self.connect_condition = Condition(self.lock)
        self.connect_attempts = 0
        self.open_listener = open_listener
        self.close_listener = close_listener
        self.fail_listener = fail_listener
        self.is_connected = False
        self.is_alive = True
        self.object_watchers = {}
        self.object_values = {}
#        net.OutputThread(socket, self.queue.get).start()
#        net.InputThread(socket, self.received, self.cleanup).start()
        # We query here so that an invalid service id will cause an exception
        # to be raised while constructing the service
        # self.query(messaging.create_command("bind", False, service=service_id), timeout=10)
        Thread(name="autobus2.remote.Connection._connect-initial", target=self._connect).start()
    
    def _connect(self):
        delay = 0.1
        delay_increment = 1.5
        delay_max = 20
        while self.is_alive:
            delay *= delay_increment
            if delay > delay_max:
                delay = delay_max
            with self.lock:
                if not self.is_alive:
                    return
            # Try to connect to the remote end
            s = Socket()
            s.settimeout(10) # TODO: make this configurable
            with self.lock:
                self.would_be_socket = s
            try:
                try:
                    s.connect((self.host, self.port))
                finally:
                    with self.lock:
                        self.would_be_socket = None
            except SocketError:
                # Connection failed; wait the specified delay, then start over
                net.shutdown(s)
                self._on_connection_failed()
                self.sleep_while_alive(delay)
                continue
            s.settimeout(None)
            # Create the queue holding messages to be sent to the remote end
            queue = Queue()
            input_thread, output_thread = net.start_io_threads(s, None, queue.get)
            # Create the queue that will be used to hold the response to the
            # initial bind command that we're going to send
            input_queue = Queue()
            # Send the initial bind command
            queue.put(messaging.create_command("bind", service=self.service_id))
            # Then register a function waiting to receive the response
            def received(message):
                # Remove ourselves so that we only receive one message, namely
                # the bind response
                input_thread.function = None
                # Then go stuff the response in the queue
                input_queue.put(message)
            input_thread.function = received
            try:
                bind_response = input_queue.get(timeout=10) # TODO: make timeout configurable
            except Empty:
                bind_response = None
            if bind_response is None or bind_response.get("_error") is not None:
                # An error happened while getting the response, so close the
                # output thread, then forcibly close the socket, then put an
                # empty function into the input thread so that it wil die if
                # any other messages were received, then sleep the amount of
                # time we're supposed to and continue
                queue.put(None)
                net.shutdown(s)
                input_thread.function = lambda *args: None
                self._on_connection_failed()
                self.sleep_while_alive(delay)
                continue
            # The connection succeeded and we've bound to the remote service
            # successfully, so we lock on ourselves and stick everything into
            # the relevant fields, and send any initial messages (initial
            # object listeners, event listeners, etc) that we need to.
            with self.lock:
                if not self.is_alive: # close() was called while we were
                    # binding, so we close everything and return
                    queue.put(None)
                    net.shutdown(s)
                    input_thread.function = lambda *args: None
                    return
                self.socket = s
                self.is_connected = True
                self.queue = queue
                input_thread.function = self.received
                self._on_connection_succeeded()
                # FIXME: send initial messages here, once we have any to send
            return
    
    def sleep_while_alive(self, amount):
        while amount > 0 and self.is_alive:
            time.sleep(amount%1.0)
            amount -= 1
    
    @synchronized_on("lock")
    def _on_connection_failed(self):
        self.connect_attempts += 1
        self.connect_condition.notify_all()
        with print_exceptions:
            if self.fail_listener:
                self.fail_listener(self)
    
    @synchronized_on("lock")
    def _on_connection_succeeded(self):
        self.connect_attempts += 1
        self.connect_condition.notify_all()
        for name in self.object_watchers:
            self.send_watch(name)
        with print_exceptions:
            if self.open_listener:
                self.open_listener(self)

    def close(self):
        with self.lock:
            if not self.is_alive: # Already closed
                return
            if self.socket:
                self.queue.put(None)
                net.shutdown(self.socket)
                self.socket = None
            if self.would_be_socket:
                net.shutdown(self.would_be_socket)
            self.is_connected = False
            self.is_alive = False
    
    def cleanup(self):
        with self.lock:
            if self.socket:
                net.shutdown(self.socket)
                self.queue.put(None)
            self.is_connected = False
            for f in self.query_map.copy().itervalues():
                with print_exceptions:
                    f(exceptions.ConnectionLostException())
            self.query_map.clear()
            for watchers in self.object_watchers.values():
                for watcher in watchers:
                    with print_exceptions:
                        watcher(None)
            with print_exceptions:
                if self.close_listener:
                    self.close_listener(self)
            if self.is_alive:
                Thread(name="autobus2.remote.Connection._connect-reconnect", target=self._connect).start()
    
    def send(self, message):
        # print "Sending: " + str(message)
        with self.lock:
            if message:
                if self.is_connected:
                    self.queue.put(message)
                else:
                    raise exceptions.NotConnectedException
    
    def send_async(self, message, callback, safe=False):
        """
        Sends the specified message. The specified callback will be called with
        the response.
        
        Unless the command isn't responded to by the remote side of the
        connection, the specified callback is guaranteed to be called. If the
        connection is not currently connected, a NotConnectedException will be
        raised. If the connection disconnects while waiting for a response, the
        callback will be called, passing in an instance of
        ConnectionLostException. If some other error occurs while processing,
        a suitable exception will be created and passed into the callback.
        
        The specified function will be called on the input thread for this
        connection, so it must not block for a significant amount of time; if
        it does, it will freeze up receiving of messages for this connection.
        
        This function (send_async) returns message["_id"]. You can later on
        pass this to cancel_query to cancel this asynchronous command if you
        want.
        
        If safe is True, send_async won't throw any exceptions, not even
        NotConnectedException or NullMessageException; these will instead
        result in the callback being synchronously called with the exception
        as its only argument.
        """
        with Consume(exceptions.AutobusException, callback, safe):
            if not message:
                raise exceptions.NullMessageException
            with self.lock:
                if not self.is_connected:
                    raise exceptions.NotConnectedException
                self.query_map[message["_id"]] = callback
                self.send(message)
    
    def query(self, message, timeout=30):
        q = Queue()
        self.send_async(message, q.put)
        try:
            response = q.get(timeout=timeout)
        except Empty: # Timeout while waiting for response
            self.cancel_query(message["_id"], False)
            raise exceptions.TimeoutException()
        if isinstance(response, Exception):
            raise response
        else:
            return response
    
    def cancel_query(self, id, call=True):
        with self.lock:
            with Suppress(KeyError): # In case the query has already finished
                function = self.query_map[id]
                if call:
                    with print_exceptions:
                        function(exceptions.QueryCanceledException())
                del self.query_map[id]
    
    @synchronized_on("lock")
    def received(self, message):
        # print "Received: " + str(message)
        if message is None: # The input thread calls this function with None as
            # the argument when the socket closes
            self.cleanup()
            return
        if message["_type"] == 2: # response
            f = self.query_map.get(message["_id"], None)
            if f:
                del self.query_map[message["_id"]]
                with print_exceptions:
                    if message.get("_error"):
                        f(exceptions.CommandErrorException(message["_error"]["text"]))
                    else:
                        f(message)
        if message["_type"] in [1, 3]:
            # TODO: consider merging this part of this class with
            # local.RemoteConnection since it's substantially the same. Maybe
            # even consider having the command/response system be its own layer
            # underneath this class, which would then function purely as the
            # layer on top of that that deals with Autobus-specific stuff.
            processor = getattr(self, "process_" + message["_command"], None)
            if not processor:
                print "Invalid message received and ignored. Command: %s" % message["_command"]
                return
            processor(message)
    
    def __getitem__(self, name):
        return Function(self, name)
    
    @synchronized_on("lock")
    def watch_object(self, name, function):
        watchers = self.object_watchers.get(name)
        if watchers is None:
            watchers = []
            self.object_watchers[name] = watchers
            if self.is_connected:
                self.send_watch(name)
        watchers.append(function)
        with print_exceptions:
            function(self.object_values.get(name))
    
    @synchronized_on("lock")
    def unwatch_object(self, name, function):
        watchers = self.object_watchers[name]
        watchers.remove(function)
        with print_exceptions:
            function(None)
        if watchers == []:
            del self.watchers[name]
            if self.is_connected:
                self.send_unwatch(name)
    
    def send_watch(self, name):
        self.send_async(messaging.create_command("watch", name=name), self.process_changed)
    
    def send_unwatch(self, name):
        self.send_async(messaging.create_command("unwatch", name=name), self.process_changed)
    
    def process_changed(self, message):
        if isinstance(message, exceptions.ConnectionLostException):
            return
        name = message["name"]
        value = message["value"]
        if value is None:
            with Suppress(KeyError):
                del self.object_values[name]
        for watcher in self.object_watchers.get(name, []):
            watcher(message["value"])
    
    def __str__(self):
        return "<%s to %s:%s service_id=%s is_connected=%s is_alive=%s>" % (
                full_name(self), self.host, self.port, self.service_id,
                self.is_connected, self.is_alive)
    
    __repr__ = __str__
    
    def wait_for_connect(self, timeout=10):
        """
        Waits until this connection has made at least one attempt since it was
        first created to connect. If the attempt succeeds, this function
        returns. If the attempt fails, this function throws an exception.
        
        If the connection's attempt to connect does not complete by the given
        timeout, an exception will be thrown.
        """
        with self.lock:
            if self.connect_attempts == 0:
                self.connect_condition.wait(timeout)
            if self.connect_attempts == 0:
                raise exceptions.TimeoutException()
            if not self.is_connected:
                raise exceptions.ConnectionException()
            


class ConnectionManager(object):
    def __init__(self, bus, connect_listener=None, disconnect_listener=None):
        self.bus = bus
    
    def connect_to(self, host, port, service_id):
        Thread(target=self._connect_to, args=(host, port, service_id)).start()
    
    def _connect_to(self, host, port, service_id):
        while True:
            try:
                connection = self.bus.connect(host, port, service_id)
            except (exceptions.ConnectionException, exceptions.TimeoutException):
                pass
            except:
                print_exc()
    
    def disconnect_from(self, host, port, service_id):
        pass
    
    @property
    def connections(self):
        """
        A list of connections that this ConnectionManager is currently
        connected to.
        """
        pass


class Function(object):
    """
    An object representing a remote function. Instances of this class are
    callable; calling them will invoke the remote function.
    
    Instances of this class can be obtained by doing
    some_connection["function_name"].
    """
    def __init__(self, connection, name):
        self.connection = connection
        self.name = name
    
    def __call__(self, *args, **kwargs):
        """
        Calls this function. The positional arguments passed to this call are
        passed to the function on the remote side when it is invoked. They must
        be JSON-encodable values; an InvalidValueException will be thrown if
        one of them is not.
        
        Three keyword arguments can be passed when calling a function:
        
        callback: This is autobus2.SYNC to call the function synchronously
        (which is what most people expect; the call will block until the remote
        service sends back a response for the function call, at which point it
        will be returned), None to call the function but return immediately
        without waiting for a response (the return value will be None), or a
        one-argument function (or other Python callable), which will cause this
        call to return immediately and the specified function to be invoked
        with the response once it arrives. The default, if callback is not
        specified, is SYNC.
        
        timeout: This only has any effect if SYNC is used as the callback (or
        the callback is not specified, since it defaults to SYNC). If the
        remote service has not sent a response in this many seconds, the call
        will stop immediately and throw a TimeoutException. The default, if
        timeout is not specified, is 30.
        
        safe: This only has any effect if a function is used as the callback.
        When safe is True, no exceptions whatsoever will be thrown; instead,
        they will be synchronously passed into the callback. If safe is False
        (the default), some exceptions (such as a function argument that can't
        be converted to JSON being passed in or the connection being currently
        disconnected) will result in an exception being immediately thrown,
        while others (such as the remote function throwing an exception, or the
        connection going down before the remote function has returned) will
        result in the callback being called with the exception as its argument.
        
        If a callback is used and an exception happens while processing (or if
        the remote function throws an exception), the exception object itself
        is passed into the callback. If a callback is not used, the exception
        will be raised instead.
        """
        callback, timeout, safe = kwargs.get("callback", autobus2.SYNC), kwargs.get("timeout", 30), kwargs.get("safe", False)
        with Consume(exceptions.AutobusException, callback, safe):
            # Make sure all the arguments can be converted into JSON correctly
            net.ensure_jsonable(args)
            # Create the command to call the function
            command = messaging.create_command("call", name=self.name, args=list(args))
            if callback is autobus2.SYNC: # Synchronous call, so we query for the
                # command
                return self.connection.query(command, timeout)["result"]
            elif callback is None:  # Asynchronous call; send the command as a
                # notice and then return
                self.connection.send(messaging.convert_to_notice(command))
            else: # Call with a callback, so we write a wrapper to handle everything
                def wrapper(response):
                    if isinstance(response, dict): # Normal response
                        callback(response["result"])
                    else: # Exception while processing
                        callback(response)
                self.connection.send_async(command, wrapper)
    
    def __str__(self):
        return "<%s %s from %s>" % (full_name(self), self.name, self.connection)
    
    __repr__ = __str__








































# This is essentially a dumping ground for various things I'm trying out that
# have to do specifically with GTK. It loads from glade/landfill/landfill.xml
# and shows the UI therein present.

import gobject
import gtk

gobject.threads_init()

checkbox_ignore = False
togglebutton_ignore = False

class Signals(object):
    def on_checkbox_clicked2(self, checkbox):
        """
        print "Checkbox clicked"
        global checkbox_ignore
        if checkbox_ignore:
            checkbox_ignore = False
            return
        checkbox_ignore = True
        checkbox.set_active(not checkbox.get_active())
        """
    
    def toggle_box_happened(self, box):
        return False
    
    def on_togglebutton_clicked2(self, togglebutton):
        print "Togglebutton clicked"
        global togglebutton_ignore
        if togglebutton_ignore:
            togglebutton_ignore = False
            return
        togglebutton_ignore = True
        togglebutton.set_active(not togglebutton.get_active())
    
    def toggle_box(self, button):
        print "Toggle checkboxbox button clicked"
        global checkbox_ignore
        checkbox_ignore = True
        checkbox = builder.get_object("checkbox")
        checkbox.set_active(not checkbox.get_active())
    
    def toggle_button(self, button):
        print "Toggle togglebutton button clicked"
        global togglebutton_ignore
        togglebutton_ignore = True
        togglebutton = builder.get_object("togglebutton")
        togglebutton.set_active(not togglebutton.get_active())
    
    def gtk_main_quit(self, *args):
        gtk.main_quit()

def main():
    global builder
    builder = gtk.Builder()
    builder.add_from_file("glade/landfill/landfill.xml")
    builder.connect_signals(Signals())
    builder.get_object("window").show()
    gtk.main()
 



import gobject
import gtk
from libautobus import AutobusConnection
from display import TimerSetDisplay

gobject.threads_init()

def main():
    bus = AutobusConnection()
    controller = bus["timer"]
    display = TimerSetDisplay("TrayTimer", controller, lambda: None)
    display.window.show()
